# Author: Nicolas Legrand <nicolas.legrand@cas.au.dk>

import os
from typing import Any, Dict, Optional

import numpy as np
import pandas as pd
import pkg_resources  # type: ignore
import serial
from systole import serialSim
from systole.recording import Oximeter


def getParameters(
    participant: str = "Participant",
    session: str = "001",
    serialPort: str = "COM3",
    taskVersion: str = "Garfinkel",
    setup: str = "behavioral",
    screenNb: int = 0,
    fullscr: bool = True,
    resultPath: Optional[str] = None,
    systole_kw: dict = {},
) -> Dict:
    """Create Heartbeat Counting task parameters.

    Parameters
    ----------
    participant : str
        Subject ID. Default is 'exteroStairCase'.
    resultPath : str or None
        Where to save the results.
    screenNb : int
        Screen number. Used to parametrize py:func:`psychopy.visual.Window`.
        Default is set to 0.
    serialPort: str
        The USB port where the pulse oximeter is plugged. Should be written as a string
        e.g. `"COM3"` for USB ports on Windows.
    session : int
        Session number. Default to '001'.
    setup : str
        Context of oximeter recording. `"behavioral"` will record through a Nonin
        pulse oximeter, `"test"` will use pre-recorded pulse time series (for testing
        only).
    systole_kw : dict
        Additional keyword arguments for :py:class:`systole.recorder.Oxmeter`.
    taskVersion : str or None
        Task version to run. Can be 'Garfinkel', 'Shandry', 'test' or None.

    Attributes
    ----------
    conditions : 1d array-like of str
        The conditions. Can be 'Rest', 'Training' or 'Count'.
    confScale : list
        The range of the confidence rating scale.
    heartLogo : `psychopy.visual.ImageStim`
        Image presented during resting conditions.
    labelsRating : list
        The labels of the confidence rating scale.
    noteStart : psychopy.sound.Sound instance
        The sound that will be played when trial starts.
    noteStop : psychopy.sound.Sound instance
        The sound that will be played when trial ends.
    path : str
        The task working directory.
    randomize : bool
        If `True` (default), will randomize the order of the conditions. If
        taskVersion is not None, will use the default task parameter instead.
    rating : bool
        If `True` (default), will add a rating scale after the evaluation.
    restLength : int
        The length of the resting period (seconds). Default is 300 seconds.
    restLogo : `psychopy.visual.ImageStim`
        Image presented during resting conditions.
    restPeriod : bool
        If `True`, a resting period will be proposed before the task.
    resultPath : str
        The subject result directory.
    screenNb : int
        The screen number (Psychopy parameter). Default set to 0.
    serial : `serial.Serial`
        The serial port used to record the PPG activity.
    startKey : str
        The key to press to start the task and go to next steps.
    taskVersion : str or None
        Task version to run. Can be 'Garfinkel', 'Shandry', 'test' or None.
    texts : dict
        Dictionary containing the texts to be presented.
    textSize : float
        Text size.
    triggers : dict
        Dictionary {str, callable or None}. The function will be executed
        before the corresponding trial sequence. The default values are
        `None` (no trigger sent).
        * `"trialStart"`
        * `"trialStop"`
        * `"listeningStart"`
        * `"listeningStop"`
        * `"decisionStart"`
        * `"decisionStop"`
        * `"confidenceStart"`
        * `"confidenceStop"`
    times : 1d array-like of int
        Length of trials, in seconds.
    win : `psychopy.visual.window`
        The window in which to draw objects.

    """
    from psychopy import sound, visual

    parameters: Dict[str, Any] = {}
    parameters["restPeriod"] = True
    parameters["restLength"] = 30
    parameters["randomize"] = True
    parameters["startKey"] = "space"
    parameters["rating"] = True
    parameters["confScale"] = [1, 7]
    parameters["labelsRating"] = ["Guess", "Certain"]
    parameters["taskVersion"] = taskVersion
    parameters["results_df"] = pd.DataFrame({})
    parameters["setup"] = setup

    # Initialize triggers dictionary with None
    # Some or all can later be overwrited with callable
    # sending the information needed.
    parameters["triggers"] = {
        "trialStart": None,
        "trialStop": None,
        "listeningStart": None,
        "listeningStop": None,
        "decisionStart": None,
        "decisionStop": None,
        "confidenceStart": None,
        "confidenceStop": None,
    }

    # Experimental design - can choose between a version based on recent
    # papers from Sarah Garfinkel's group, or the classic Schandry approach.
    # The primary difference ebtween the two is the order of trials and the
    # use of resting periods between trials.
    if parameters["taskVersion"] == "Garfinkel":
        parameters["times"] = np.array([25, 30, 35, 40, 45, 50])
        np.random.shuffle(parameters["times"])
        parameters["conditions"] = [
            "Count",
            "Count",
            "Count",
            "Count",
            "Count",
            "Count",
        ]

    elif parameters["taskVersion"] == "Schandry":
        parameters["times"] = np.array([60, 25, 30, 35, 30, 45])
        parameters["conditions"] = ["Rest", "Count", "Rest", "Count", "Rest", "Count"]

    elif parameters["taskVersion"] == "test":
        parameters["times"] = np.array([5, 5])
        parameters["conditions"] = ["Rest", "Count"]
    else:
        raise ValueError("Invalid task condition")

    # Set default path /Results/ 'Subject ID' /
    parameters["participant"] = participant
    parameters["session"] = session
    parameters["path"] = os.getcwd()
    if resultPath is None:
        parameters["resultPath"] = parameters["path"] + "/data/" + participant + session
    else:
        parameters["resultPath"] = resultPath
    # Create Results directory of not already exists
    if not os.path.exists(parameters["resultPath"]):
        os.makedirs(parameters["resultPath"])

    # Set note played at trial start
    parameters["noteStart"] = sound.Sound(
        pkg_resources.resource_filename("cardioception.HBC", "Sounds/start.wav")
    )

    parameters["noteStop"] = sound.Sound(
        pkg_resources.resource_filename("cardioception.HBC", "Sounds/stop.wav")
    )

    # Open window
    if parameters["setup"] == "test":
        fullscr = False
    parameters["win"] = visual.Window(screen=screenNb, fullscr=fullscr, units="height")
    parameters["win"].mouseVisible = False

    parameters["restLogo"] = visual.ImageStim(
        win=parameters["win"],
        units="height",
        image=pkg_resources.resource_filename(__name__, "Images/rest.png"),
        pos=(0.0, -0.2),
    )
    parameters["restLogo"].size *= 0.15
    parameters["heartLogo"] = visual.ImageStim(
        win=parameters["win"],
        units="height",
        image=pkg_resources.resource_filename(__name__, "Images/heartbeat.png"),
        pos=(0.0, -0.2),
    )
    parameters["heartLogo"].size *= 0.05

    if setup == "behavioral":
        # PPG recording
        port = serial.Serial(serialPort)
        parameters["oxiTask"] = Oximeter(
            serial=port, sfreq=75, add_channels=1, **systole_kw
        )
        parameters["oxiTask"].setup().read(duration=1)
    elif setup == "test":
        # Use pre-recorded pulse time series for testing
        port = serialSim()
        parameters["oxiTask"] = Oximeter(
            serial=port, sfreq=75, add_channels=1, **systole_kw
        )
        parameters["oxiTask"].setup().read(duration=1)

    #######
    # Texts
    #######

    # Task instructions
    parameters["texts"] = dict()
    parameters["texts"]["Rest"] = "Please sit quietly until the next session"
    parameters["texts"]["Count"] = (
        "After you hear START, try to count your heartbeats"
        " by concentrating on your body feelings."
        " Stop counting when you hear STOP"
    )
    parameters["texts"]["Training"] = (
        "After you hear START, try to count your heartbeats"
        " by concentrating on your body feelings"
        " Stop counting when you hear STOP"
    )
    parameters["texts"]["nCount"] = (
        "How many heartbeats did you count?"
        " Write a number and press ENTER to validate."
    )
    parameters["texts"]["confidence"] = (
        "How confident are you about your count?"
        "Use the RIGHT/LEFT keys to select and the DOWN key to confirm"
    )

    # Tutorial instructions
    parameters["texts"]["Tutorial1"] = (
        "During this experiment, we will ask you to silently"
        " count your heartbeats for different intervals of time."
    )
    parameters["texts"]["Tutorial2"] = (
        'When you see this "heart" icon, you will silently count your'
        " heartbeats by focusing on your body sensations."
    )
    parameters["texts"]["Tutorial3"] = (
        'Sometime, you will also encounter this "rest" icon.'
        " In this case your task will just be to sit quietly until the next"
        " session."
    )
    parameters["texts"]["Tutorial4"] = (
        "The beginning and the end of the task will be signalled when you hear"
        " the words 'START'' and 'STOP'. While counting your heartbeats, you"
        " may close your eyes if you find that helpful. Please keep your hand"
        " still during the counting period, to avoid interfering with"
        " the heartbeat recording."
    )
    parameters["texts"]["Tutorial5"] = (
        "After the counting part of the task, you will be asked to report the"
        " exact number of heartbeats you felt during the interval between"
        " 'START' and 'STOP'. Please do not try to estimate the number of"
        " heartbeats, but instead only report the heartbeats you actually felt"
        " during the interval. You will input your response using the number"
        " pad and press return when done. You can also correct your response"
        " using backspace."
    )
    parameters["texts"]["Tutorial6"] = (
        "Once you have made your response, you will estimate your subjective"
        " feeling of confidence in how accurate your count was"
        " for that interval. A large number here means that you are totally"
        " certain you counted the exact number of heartbeats that occured,"
        " and a small number means that you are totally uncertain or felt that"
        " you were guessing about the"
        " number of heartbeats. You should use the RIGHT and LEFT"
        " key to select your response and the DOWN key to confirm."
    )
    parameters["texts"]["Tutorial7"] = (
        "Before the main task begins there is a short resting period of"
        " several minutes, during which we will calibrate the heartbeat"
        " recording. During this period, please sit quietly with your"
        " hands still to avoid interfering with the calibration."
        " Afterwards, the counting task will begin, and will take about"
        " 6 minutes in total."
    )
    parameters["texts"]["Tutorial8"] = (
        "You will now complete a short practice task."
        " Please ask the experimenter if you have any questions before"
        " continuing to the main experiment."
    )
    parameters["texts"]["Tutorial9"] = (
        "Good job! If you have any question, ask the experimenter now,"
        " otherwise press SPACE to continue to the experiment."
    )
    parameters["textSize"] = 0.04

    return parameters



# Author: Nicolas Legrand <nicolas.legrand@cas.au.dk>

import pickle
import time
from typing import Optional, Tuple

import numpy as np
import pandas as pd
import pkg_resources  # type: ignore
from systole.detection import ppg_peaks


def run(
    parameters: dict,
    confidenceRating: bool = True,
    runTutorial: bool = False,
):
    """Run the Heart Rate Discrimination task.

    Parameters
    ----------
    parameters : dict
        Task parameters.
    confidenceRating : bool
        Whether the trial show include a confidence rating scale.
    runTutorial : bool
        If `True`, will present a tutorial with 10 training trial with feedback
        and 5 trials with confidence rating.
    """
    from psychopy import core, visual

    # Initialization of the Pulse Oximeter
    parameters["oxiTask"].setup().read(duration=1)

    # Show tutorial and training trials
    if runTutorial is True:
        tutorial(parameters)

    for nTrial, modality, trialType in zip(
        range(parameters["nTrials"]),
        parameters["Modality"],
        parameters["staircaseType"],
    ):
        # Initialize variable
        estimatedThreshold, estimatedSlope = None, None

        # Wait for key press if this is the first trial
        if nTrial == 0:
            # Ask the participant to press default button to start
            messageStart = visual.TextStim(
                parameters["win"],
                height=parameters["textSize"],
                text=parameters["texts"]["textTaskStart"],
            )
            press = visual.TextStim(
                parameters["win"],
                height=parameters["textSize"],
                pos=(0.0, -0.4),
                text=parameters["texts"]["textNext"],
            )
            press.draw()
            messageStart.draw()  # Show instructions
            parameters["win"].flip()

            waitInput(parameters)

        # Next intensity value
        if trialType == "updown":
            print("... load UpDown staircase.")
            thisTrial = parameters["stairCase"][modality].next()
            stairCond = thisTrial[1]["label"]
            alpha = thisTrial[0]
        elif trialType == "psi":
            print("... load psi staircase.")
            alpha = parameters["stairCase"][modality].next()
            stairCond = "psi"
        elif trialType == "CatchTrial":
            print("... load catch trial.")
            # Select pseudo-random extrem value based on number
            # of previous catch trial.
            catchIdx = sum(
                parameters["staircaseType"][:nTrial][
                    parameters["Modality"][:nTrial] == modality
                ]
                == "CatchTrial"
            )
            alpha = np.array([-30, 10, -20, 20, -10, 30])[catchIdx % 6]
            stairCond = "CatchTrial"

        # Before trial triggers
        parameters["oxiTask"].readInWaiting()
        parameters["oxiTask"].channels["Channel_0"][-1] = 1  # Trigger

        # Start trial
        (
            condition,
            listenBPM,
            responseBPM,
            decision,
            decisionRT,
            confidence,
            confidenceRT,
            alpha,
            is_correct,
            response_provided,
            ratingProvided,
            startTrigger,
            soundTrigger,
            responseMadeTrigger,
            ratingStartTrigger,
            ratingEndTrigger,
            endTrigger,
        ) = trial(
            parameters,
            alpha,
            modality,
            confidenceRating=confidenceRating,
            nTrial=nTrial,
        )

        # Check if response is 'More' or 'Less'
        isMore = 1 if decision == "More" else 0
        # Update the UpDown staircase if initialization trial
        if trialType == "updown":
            print("... update UpDown staircase.")
            # Update the UpDown staircase
            parameters["stairCase"][modality].addResponse(isMore)
        elif trialType == "psi":
            print("... update psi staircase.")

            # Update the Psi staircase with forced intensity value
            # if impossible BPM was generated
            if listenBPM + alpha < 15:
                parameters["stairCase"][modality].addResponse(isMore, intensity=15)
            elif listenBPM + alpha > 199:
                parameters["stairCase"][modality].addResponse(isMore, intensity=199)
            else:
                parameters["stairCase"][modality].addResponse(isMore)

            # Store posteriors in list for each trials
            parameters["staircaisePosteriors"][modality].append(
                parameters["stairCase"][modality]._psi._probLambda[0, :, :, 0]
            )

            # Save estimated threshold and slope for each trials
            estimatedThreshold, estimatedSlope = parameters["stairCase"][
                modality
            ].estimateLambda()

        print(
            f"... Initial BPM: {listenBPM} - Staircase value: {alpha} "
            f"- Response: {decision} ({is_correct})"
        )

        # Store results
        parameters["results_df"] = pd.concat(
            [
                parameters["results_df"],
                pd.DataFrame(
                    {
                        "TrialType": [trialType],
                        "Condition": [condition],
                        "Modality": [modality],
                        "StairCond": [stairCond],
                        "Decision": [decision],
                        "DecisionRT": [decisionRT],
                        "Confidence": [confidence],
                        "ConfidenceRT": [confidenceRT],
                        "Alpha": [alpha],
                        "listenBPM": [listenBPM],
                        "responseBPM": [responseBPM],
                        "ResponseCorrect": [is_correct],
                        "DecisionProvided": [response_provided],
                        "RatingProvided": [ratingProvided],
                        "nTrials": [nTrial],
                        "EstimatedThreshold": [estimatedThreshold],
                        "EstimatedSlope": [estimatedSlope],
                        "StartListening": [startTrigger],
                        "StartDecision": [soundTrigger],
                        "ResponseMade": [responseMadeTrigger],
                        "RatingStart": [ratingStartTrigger],
                        "RatingEnds": [ratingEndTrigger],
                        "endTrigger": [endTrigger],
                    }
                ),
            ],
            ignore_index=True,
        )

        # Save the results at each iteration
        parameters["results_df"].to_csv(
            parameters["resultPath"]
            + "/"
            + parameters["participant"]
            + parameters["session"]
            + ".txt",
            index=False,
        )

        # Breaks
        if (nTrial % parameters["nBreaking"] == 0) & (nTrial != 0):
            message = visual.TextStim(
                parameters["win"],
                height=parameters["textSize"],
                text=parameters["texts"]["textBreaks"],
            )
            percRemain = round((nTrial / parameters["nTrials"]) * 100, 2)
            remain = visual.TextStim(
                parameters["win"],
                height=parameters["textSize"],
                pos=(0.0, 0.2),
                text=f" ---- {percRemain} % ---- ",
            )
            remain.draw()
            message.draw()
            parameters["win"].flip()
            parameters["oxiTask"].save(
                f"{parameters['resultPath']}/{parameters['participant']}_ppg_{nTrial}.txt"
            )

            # Wait for participant input before continue
            waitInput(parameters)

            # Fixation cross
            fixation = visual.GratingStim(
                win=parameters["win"], mask="cross", size=0.1, pos=[0, 0], sf=0
            )
            fixation.draw()
            parameters["win"].flip()

            # Reset recording when ready
            parameters["oxiTask"].setup()
            parameters["oxiTask"].read(duration=1)

    # Save the final results
    print("Saving final results in .txt file...")
    parameters["results_df"].to_csv(
        parameters["resultPath"]
        + "/"
        + parameters["participant"]
        + parameters["session"]
        + "_final.txt",
        index=False,
    )

    # Save the final signals file
    print("Saving PPG signal data frame...")
    parameters["signal_df"].to_csv(
        parameters["resultPath"] + "/" + parameters["participant"] + "_signal.txt",
        index=False,
    )

    # Save last pulse oximeter recording, if relevant
    parameters["oxiTask"].save(
        f"{parameters['resultPath']}/{parameters['participant']}_ppg_{nTrial}_end.txt"
    )

    # Save posterios (if relevant)
    print("Saving posterior distributions...")
    for k in set(parameters["Modality"]):
        np.save(
            parameters["resultPath"]
            + "/"
            + parameters["participant"]
            + k
            + "_posterior.npy",
            np.array(parameters["staircaisePosteriors"][k]),
        )

    # Save parameters
    print("Saving Parameters in pickle...")
    save_parameter = parameters.copy()
    for k in ["win", "heartLogo", "listenLogo", "stairCase", "oxiTask"]:
        del save_parameter[k]
    if parameters["device"] == "mouse":
        del save_parameter["myMouse"]
    del save_parameter["handSchema"]
    del save_parameter["pulseSchema"]
    with open(
        save_parameter["resultPath"]
        + "/"
        + save_parameter["participant"]
        + "_parameters.pickle",
        "wb",
    ) as handle:
        pickle.dump(save_parameter, handle, protocol=pickle.HIGHEST_PROTOCOL)

    # End of the task
    end = visual.TextStim(
        parameters["win"],
        height=parameters["textSize"],
        pos=(0.0, 0.0),
        text=parameters["texts"]["done"],
    )
    end.draw()
    parameters["win"].flip()
    core.wait(3)


def trial(
    parameters: dict,
    alpha: float,
    modality: str,
    confidenceRating: bool = True,
    feedback: bool = False,
    nTrial: Optional[int] = None,
) -> Tuple[
    str,
    float,
    float,
    Optional[str],
    Optional[float],
    Optional[float],
    Optional[float],
    float,
    Optional[bool],
    bool,
    bool,
    float,
    float,
    float,
    Optional[float],
    Optional[float],
    float,
]:
    """Run one trial of the Heart Rate Discrimination task.

    Parameters
    ----------
    parameter : dict
        Task parameters.
    alpha : float
        The intensity of the stimulus, from the staircase procedure.
    modality : str
        The modality, can be `'Intero'` or `'Extro'` if an exteroceptive
        control condition has been added.
    confidenceRating : boolean
        If `False`, do not display confidence rating scale.
    feedback : boolean
        If `True`, will provide feedback.
    nTrial : int
        Trial number (optional).

    Returns
    -------
    condition : str
        The trial condition, can be `'Higher'` or `'Lower'` depending on the
        alpha value.
    listenBPM : float
        The frequency of the tones (exteroceptive condition) or of the heart
        rate (interoceptive condition), expressed in BPM.
    responseBPM : float
        The frequency of thefeebdack tones, expressed in BPM.
    decision : str
        The participant decision. Can be `'up'` (the participant indicates
        the beats are faster than the recorded heart rate) or `'down'` (the
        participant indicates the beats are slower than recorded heart rate).
    decisionRT : float
        The response time from sound start to choice (seconds).
    confidence : int
        If confidenceRating is *True*, the confidence of the participant. The
        range of the scale is defined in `parameters['confScale']`. Default is
        `[1, 7]`.
    confidenceRT : float
        The response time (RT) for the confidence rating scale.
    alpha : int
        The difference between the true heart rate and the delivered tone BPM.
        Alpha is defined by the stairCase.intensities values and is updated
        on each trial.
    is_correct : int
        `0` for incorrect response, `1` for correct responses. Note that this
        value is not feeded to the staircase when using the (Yes/No) version
        of the task, but instead will check if the response is `'More'` or not.
    response_provided : bool
        Was the decision provided (`True`) or not (`False`).
    ratingProvided : bool
        Was the rating provided (`True`) or not (`False`). If no decision was
        provided, the ratig scale is not proposed and no ratings can be provided.
    startTrigger, soundTrigger, responseMadeTrigger, ratingStartTrigger,\
        ratingEndTrigger, endTrigger : float
        Time stamp of key timepoints inside the trial.
    """
    from psychopy import core, event, sound, visual

    # Print infos at each trial start
    print(f"Starting trial - Intensity: {alpha} - Modality: {modality}")

    parameters["win"].mouseVisible = False

    # Restart the trial until participant provide response on time
    confidence, confidenceRT, is_correct, ratingProvided = None, None, None, False

    # Fixation cross
    fixation = visual.GratingStim(
        win=parameters["win"], mask="cross", size=0.1, pos=[0, 0], sf=0
    )
    fixation.draw()
    parameters["win"].flip()
    core.wait(np.random.uniform(parameters["isi"][0], parameters["isi"][1]))

    keys = event.getKeys()
    if "escape" in keys:
        print("User abort")
        parameters["win"].close()
        core.quit()

    if modality == "Intero":
        ###########
        # Recording
        ###########
        messageRecord = visual.TextStim(
            parameters["win"],
            height=parameters["textSize"],
            pos=(0.0, 0.2),
            text=parameters["texts"]["textHeartListening"],
        )
        messageRecord.draw()

        # Start recording trigger
        parameters["oxiTask"].readInWaiting()
        parameters["oxiTask"].channels["Channel_0"][-1] = 2  # Trigger

        parameters["heartLogo"].draw()
        parameters["win"].flip()

        startTrigger = time.time()

        # Recording
        while True:
            # Read the raw PPG signal from the pulse oximeter
            # You can adapt these line to work with a different setup provided that
            # it can measure and create the new variable `bpm` (the average beats per
            # minute over the 5 seconds of recording).
            signal = (
                parameters["oxiTask"].read(duration=5.0).recording[-75 * 6 :]  # noqa
            )
            signal, peaks = ppg_peaks(signal, sfreq=75, new_sfreq=1000, clipping=True)

            # Get actual heart Rate
            # Only use the last 5 seconds of the recording
            bpm = 60000 / np.diff(np.where(peaks[-5000:])[0])

            print(f"... bpm: {[round(i) for i in bpm]}")

            # Prevent crash if NaN value
            if np.isnan(bpm).any() or (bpm is None) or (bpm.size == 0):
                message = visual.TextStim(
                    parameters["win"],
                    height=parameters["textSize"],
                    text=parameters["texts"]["checkOximeter"],
                    color="red",
                )
                message.draw()
                parameters["win"].flip()
                core.wait(2)

            else:
                # Check for extreme heart rate values, if crosses theshold,
                # hold the task until resolved. Cutoff values determined in
                # parameters to correspond to biologically unlikely values.
                if not (
                    (np.any(bpm < parameters["HRcutOff"][0]))
                    or (np.any(bpm > parameters["HRcutOff"][1]))
                ):
                    listenBPM = round(bpm.mean() * 2) / 2  # Round nearest .5
                    break
                else:
                    message = visual.TextStim(
                        parameters["win"],
                        height=parameters["textSize"],
                        text=parameters["texts"]["stayStill"],
                        color="red",
                    )
                    message.draw()
                    parameters["win"].flip()
                    core.wait(2)

    elif modality == "Extero":
        ###########
        # Recording
        ###########
        messageRecord = visual.TextStim(
            parameters["win"],
            height=parameters["textSize"],
            pos=(0.0, 0.2),
            text=parameters["texts"]["textToneListening"],
        )
        messageRecord.draw()

        # Start recording trigger
        parameters["oxiTask"].readInWaiting()
        parameters["oxiTask"].channels["Channel_0"][-1] = 2  # Trigger

        parameters["listenLogo"].draw()
        parameters["win"].flip()

        startTrigger = time.time()

        # Random selection of HR frequency
        listenBPM = np.random.choice(np.arange(40, 100, 0.5))

        # Play the corresponding beat file
        listenFile = pkg_resources.resource_filename(
            "cardioception.HRD", f"Sounds/{listenBPM}.wav"
        )
        print(f"...loading file (Listen): {listenFile}")

        # Play selected BPM frequency
        listenSound = sound.Sound(listenFile)
        listenSound.play()
        core.wait(5)
        listenSound.stop()

    else:
        raise ValueError("Invalid modality")

    # Fixation cross
    fixation = visual.GratingStim(
        win=parameters["win"], mask="cross", size=0.1, pos=[0, 0], sf=0
    )
    fixation.draw()
    parameters["win"].flip()
    core.wait(0.5)

    #######
    # Sound
    #######

    # Generate actual stimulus frequency
    condition = "Less" if alpha < 0 else "More"

    # Check for extreme alpha values, e.g. if alpha changes massively from
    # trial to trial.
    if (listenBPM + alpha) < 15:
        responseBPM = 15.0
    elif (listenBPM + alpha) > 199:
        responseBPM = 199.0
    else:
        responseBPM = listenBPM + alpha
    responseFile = pkg_resources.resource_filename(
        "cardioception.HRD", f"Sounds/{responseBPM}.wav"
    )
    print(f"...loading file (Response): {responseFile}")

    # Play selected BPM frequency
    responseSound = sound.Sound(responseFile)
    if modality == "Intero":
        parameters["heartLogo"].autoDraw = True
    elif modality == "Extero":
        parameters["listenLogo"].autoDraw = True
    else:
        raise ValueError("Invalid modality provided")
    # Record participant response (+/-)
    message = visual.TextStim(
        parameters["win"],
        height=parameters["textSize"],
        pos=(0, 0.4),
        text=parameters["texts"]["Decision"][modality],
    )
    message.autoDraw = True

    press = visual.TextStim(
        parameters["win"],
        height=parameters["textSize"],
        text=parameters["texts"]["responseText"],
        pos=(0.0, -0.4),
    )
    press.autoDraw = True

    # Sound trigger
    parameters["oxiTask"].readInWaiting()
    parameters["oxiTask"].channels["Channel_0"][-1] = 3
    soundTrigger = time.time()
    parameters["win"].flip()

    #####################
    # Esimation Responses
    #####################
    (
        responseMadeTrigger,
        response_trigger,
        response_provided,
        decision,
        decisionRT,
        is_correct,
    ) = responseDecision(responseSound, parameters, feedback, condition)
    press.autoDraw = False
    message.autoDraw = False
    if modality == "Intero":
        parameters["heartLogo"].autoDraw = False
    elif modality == "Extero":
        parameters["listenLogo"].autoDraw = False
    else:
        raise ValueError("Invalid modality provided")
    ###################
    # Confidence Rating
    ###################

    # Record participant confidence
    if (confidenceRating is True) & (response_provided is True):
        # Confidence rating start trigger
        parameters["oxiTask"].readInWaiting()
        parameters["oxiTask"].channels["Channel_0"][-1] = 4  # Trigger

        # Confidence rating scale
        ratingStartTrigger: Optional[float] = time.time()
        (
            confidence,
            confidenceRT,
            ratingProvided,
            ratingEndTrigger,
        ) = confidenceRatingTask(parameters)
    else:
        ratingStartTrigger, ratingEndTrigger = None, None

    # Confidence rating end trigger
    parameters["oxiTask"].readInWaiting()
    parameters["oxiTask"].channels["Channel_0"][-1] = 5
    endTrigger = time.time()

    # Save PPG signal
    if nTrial is not None:  # Not during the tutorial
        if modality == "Intero":
            this_df = None
            # Save physio signal
            this_df = pd.DataFrame(
                {
                    "signal": signal,
                    "nTrial": pd.Series([nTrial] * len(signal), dtype="category"),
                }
            )

            parameters["signal_df"] = pd.concat(
                [parameters["signal_df"], this_df], ignore_index=True
            )

    return (
        condition,
        listenBPM,
        responseBPM,
        decision,
        decisionRT,
        confidence,
        confidenceRT,
        alpha,
        is_correct,
        response_provided,
        ratingProvided,
        startTrigger,
        soundTrigger,
        responseMadeTrigger,
        ratingStartTrigger,
        ratingEndTrigger,
        endTrigger,
    )


def waitInput(parameters: dict):
    """Wait for participant input before continue"""

    from psychopy import core, event

    if parameters["device"] == "keyboard":
        while True:
            keys = event.getKeys()
            if "escape" in keys:
                print("User abort")
                parameters["win"].close()
                core.quit()
            elif parameters["startKey"] in keys:
                break
    elif parameters["device"] == "mouse":
        parameters["myMouse"].clickReset()
        while True:
            buttons = parameters["myMouse"].getPressed()
            if buttons != [0, 0, 0]:
                break
            keys = event.getKeys()
            if "escape" in keys:
                print("User abort")
                parameters["win"].close()
                core.quit()


def tutorial(parameters: dict):
    """Run tutorial before task run.

    Parameters
    ----------
    parameters : dict
        Task parameters.

    """

    from psychopy import core, event, visual

    # Introduction
    intro = visual.TextStim(
        parameters["win"],
        height=parameters["textSize"],
        text=parameters["texts"]["Tutorial1"],
    )
    press = visual.TextStim(
        parameters["win"],
        height=parameters["textSize"],
        pos=(0.0, -0.4),
        text=parameters["texts"]["textNext"],
    )
    intro.draw()
    press.draw()
    parameters["win"].flip()
    core.wait(1)

    waitInput(parameters)

    # Pusle oximeter tutorial
    pulse1 = visual.TextStim(
        parameters["win"],
        height=parameters["textSize"],
        pos=(0.0, 0.3),
        text=parameters["texts"]["pulseTutorial1"],
    )
    press = visual.TextStim(
        parameters["win"],
        height=parameters["textSize"],
        pos=(0.0, -0.4),
        text=parameters["texts"]["textNext"],
    )
    pulse1.draw()
    parameters["pulseSchema"].draw()
    press.draw()
    parameters["win"].flip()
    core.wait(1)

    waitInput(parameters)

    # Get finger number - Skip this part for the danish_children version (empty string)
    if parameters["texts"]["pulseTutorial2"]:
        pulse2 = visual.TextStim(
            parameters["win"],
            height=parameters["textSize"],
            pos=(0.0, 0.2),
            text=parameters["texts"]["pulseTutorial2"],
        )
        pulse3 = visual.TextStim(
            parameters["win"],
            height=parameters["textSize"],
            pos=(0.0, -0.2),
            text=parameters["texts"]["pulseTutorial3"],
        )
        pulse2.draw()
        pulse3.draw()
        press.draw()
        parameters["win"].flip()
        core.wait(1)

        waitInput(parameters)

    pulse4 = visual.TextStim(
        parameters["win"],
        height=parameters["textSize"],
        pos=(0.0, 0.3),
        text=parameters["texts"]["pulseTutorial4"],
    )
    pulse4.draw()
    parameters["handSchema"].draw()
    parameters["win"].flip()
    core.wait(1)

    # Record number
    nFinger = ""
    while True:
        # Record new key
        key = event.waitKeys(
            keyList=[
                "1",
                "2",
                "3",
                "4",
                "5",
                "num_1",
                "num_2",
                "num_3",
                "num_4",
                "num_5",
            ]
        )
        if key:
            nFinger += [s for s in key[0] if s.isdigit()][0]

            # Save the finger number in the task parameters dictionary
            parameters["nFinger"] = nFinger

            core.wait(0.5)
            break

    # Heartrate recording
    recording = visual.TextStim(
        parameters["win"],
        height=parameters["textSize"],
        pos=(0.0, 0.3),
        text=parameters["texts"]["Tutorial2"],
    )
    recording.draw()
    parameters["heartLogo"].draw()
    press.draw()
    parameters["win"].flip()
    core.wait(1)

    waitInput(parameters)

    # Show reponse icon
    listenIcon = visual.TextStim(
        parameters["win"],
        height=parameters["textSize"],
        pos=(0.0, 0.3),
        text=parameters["texts"]["Tutorial3_icon"],
    )
    parameters["heartLogo"].draw()
    listenIcon.draw()
    press.draw()
    parameters["win"].flip()
    core.wait(1)

    waitInput(parameters)

    # Response instructions
    listenResponse = visual.TextStim(
        parameters["win"],
        height=parameters["textSize"],
        pos=(0.0, 0.0),
        text=parameters["texts"]["Tutorial3_responses"],
    )
    listenResponse.draw()
    press.draw()
    parameters["win"].flip()
    core.wait(1)

    waitInput(parameters)

    # Run training trials with feedback
    parameters["oxiTask"].setup().read(duration=2)
    for i in range(parameters["nFeedback"]):
        # Ramdom selection of condition
        condition = np.random.choice(["More", "Less"])
        alpha = -20.0 if condition == "Less" else 20.0

        _ = trial(
            parameters,
            alpha,
            "Intero",
            feedback=True,
            confidenceRating=False,
        )

    # If extero conditions required, show tutorial.
    if parameters["ExteroCondition"] is True:
        exteroText = visual.TextStim(
            parameters["win"],
            height=parameters["textSize"],
            pos=(0.0, -0.2),
            text=parameters["texts"]["Tutorial3bis"],
        )
        exteroText.draw()
        parameters["listenLogo"].draw()
        press.draw()
        parameters["win"].flip()
        core.wait(1)

        waitInput(parameters)

        exteroResponse = visual.TextStim(
            parameters["win"],
            height=parameters["textSize"],
            pos=(0.0, 0.0),
            text=parameters["texts"]["Tutorial3ter"],
        )
        exteroResponse.draw()
        press.draw()
        parameters["win"].flip()
        core.wait(1)

        waitInput(parameters)

        # Run 10 training trials with feedback
        parameters["oxiTask"].setup().read(duration=2)
        for i in range(parameters["nFeedback"]):
            # Ramdom selection of condition
            condition = np.random.choice(["More", "Less"])
            alpha = -20.0 if condition == "Less" else 20.0

            _ = trial(
                parameters,
                alpha,
                "Extero",
                feedback=True,
                confidenceRating=False,
            )

    ###################
    # Confidence rating
    ###################
    confidenceText = visual.TextStim(
        parameters["win"],
        height=parameters["textSize"],
        text=parameters["texts"]["Tutorial4"],
    )
    confidenceText.draw()
    press.draw()
    parameters["win"].flip()
    core.wait(1)

    waitInput(parameters)

    parameters["oxiTask"].setup().read(duration=2)

    # Run n training trials with confidence rating
    for i in range(parameters["nConfidence"]):
        modality = "Intero"
        condition = np.random.choice(["More", "Less"])
        stim_intense = np.random.choice(np.array([1, 10, 30]))
        alpha = -stim_intense if condition == "Less" else stim_intense
        _ = trial(parameters, alpha, modality, confidenceRating=True)

    # If extero conditions required, show tutorial.
    if parameters["ExteroCondition"] is True:
        # Run n training trials with confidence rating
        for i in range(parameters["nConfidence"]):
            modality = "Extero"
            condition = np.random.choice(["More", "Less"])
            stim_intense = np.random.choice(np.array([1, 10, 30]))
            alpha = -stim_intense if condition == "Less" else stim_intense
            _ = trial(
                parameters,
                alpha,
                modality,
                confidenceRating=True,
            )

    #################
    # End of tutorial
    #################
    taskPresentation = visual.TextStim(
        parameters["win"],
        height=parameters["textSize"],
        text=parameters["texts"]["Tutorial5"],
    )
    taskPresentation.draw()
    press.draw()
    parameters["win"].flip()
    core.wait(1)
    waitInput(parameters)

    # Task
    taskPresentation = visual.TextStim(
        parameters["win"],
        height=parameters["textSize"],
        text=parameters["texts"]["Tutorial6"],
    )
    taskPresentation.draw()
    press.draw()
    parameters["win"].flip()
    core.wait(1)
    waitInput(parameters)


def responseDecision(
    this_hr,
    parameters: dict,
    feedback: bool,
    condition: str,
) -> Tuple[
    float, Optional[float], bool, Optional[str], Optional[float], Optional[bool]
]:
    """Recording response during the decision phase.

    Parameters
    ----------
    this_hr : psychopy sound instance
        The sound .wav file to play.
    parameters : dict
        Parameters dictionary.
    feedback : bool
        If `True`, provide feedback after decision.
    condition : str
        The trial condition [`'More'` or `'Less'`] used to check is response is
        correct or not.

    Returns
    -------
    responseMadeTrigger : float
        Time stamp of response provided.
    response_trigger : float
        Time stamp of response start.
    response_provided : bool
        `True` if the response was provided, `False` otherwise.
    decision : str or None
        The decision made ('Higher', 'Lower' or None)
    decisionRT : float
        Decision response time (seconds).
    is_correct : bool or None
        `True` if the response provided was correct, `False` otherwise.

    """

    from psychopy import core, event, visual

    print("...starting decision phase.")

    decision, decisionRT, is_correct = None, None, None
    response_trigger = time.time()

    if parameters["device"] == "keyboard":

        # play the tones and record key press with time stamp
        this_hr.play()
        clock = core.Clock()
        response_key = event.waitKeys(
            keyList=[
                parameters["response_keys"]["More"], 
                parameters["response_keys"]["Less"]
                ],
            maxWait=parameters["respMax"],
            timeStamped=clock,
        )
        this_hr.stop()
        responseMadeTrigger = time.time()

        # Check if the response was provided by the participant and log responses
        if not response_key:
            response_provided = False
            decision, decisionRT = None, None
        else:
            response_provided = True
            decision = response_key[0][0]
            decisionRT = response_key[0][1]

            # Is the answer Correct?
            is_correct = decision == parameters["response_keys"][condition]

    elif parameters["device"] == "mouse":
        # Initialise response feedback
        slower = visual.TextStim(
            parameters["win"],
            height=parameters["textSize"],
            color="white",
            text=parameters["texts"]["slower"],
            pos=(-0.2, 0.2),
        )
        faster = visual.TextStim(
            parameters["win"],
            height=parameters["textSize"],
            color="white",
            text=parameters["texts"]["faster"],
            pos=(0.2, 0.2),
        )
        slower.draw()
        faster.draw()
        parameters["win"].flip()

        this_hr.play()
        clock = core.Clock()
        clock.reset()
        parameters["myMouse"].clickReset()
        buttons, decisionRT = parameters["myMouse"].getPressed(getTime=True)
        while True:
            buttons, decisionRT = parameters["myMouse"].getPressed(getTime=True)
            trialdur = clock.getTime()
            parameters["oxiTask"].readInWaiting()
            if buttons == [1, 0, 0]:
                decisionRT = decisionRT[0]
                decision, response_provided = "Less", True
                slower.color = "blue"
                slower.draw()
                parameters["win"].flip()

                # Show feedback for .5 seconds if enough time
                remain = parameters["respMax"] - trialdur
                pauseFeedback = 0.5 if (remain > 0.5) else remain
                core.wait(pauseFeedback)
                break
            elif buttons == [0, 0, 1]:
                decisionRT = decisionRT[-1]
                decision, response_provided = "More", True
                faster.color = "blue"
                faster.draw()
                parameters["win"].flip()

                # Show feedback for .5 seconds if enough time
                remain = parameters["respMax"] - trialdur
                pauseFeedback = 0.5 if (remain > 0.5) else remain
                core.wait(pauseFeedback)
                break
            elif trialdur > parameters["respMax"]:  # if too long
                response_provided = False
                decisionRT = None
                break
            else:
                slower.draw()
                faster.draw()
                parameters["win"].flip()
        responseMadeTrigger = time.time()
        this_hr.stop()

        # Is the answer Correct?
        is_correct = True if (decision == condition) else False

    # Check for response provided by the participant and send feedback
    # This part is common to the mouse and keyboard versions
    if response_provided is False:
        # Record participant response (+/-)
        message = visual.TextStim(
            parameters["win"],
            height=parameters["textSize"],
            text=parameters["texts"]["tooLate"],
            color="red",
            pos=(0.0, -0.2),
        )
        message.draw()
        parameters["win"].flip()
        core.wait(0.5)

        # Read oximeter
        parameters["oxiTask"].readInWaiting()
    else:
        # Feedback
        if feedback is True:
            if is_correct == 0:
                textFeedback = parameters["texts"]["incorrectResponse"]
            else:
                textFeedback = parameters["texts"]["correctResponse"]
            colorFeedback = "red" if is_correct == 0 else "green"
            acc = visual.TextStim(
                parameters["win"],
                height=parameters["textSize"],
                pos=(0.0, -0.2),
                color=colorFeedback,
                text=textFeedback,
            )
            acc.draw()
            parameters["win"].flip()
            core.wait(1)

            # Read oximeter
            parameters["oxiTask"].readInWaiting()

    return (
        responseMadeTrigger,
        response_trigger,
        response_provided,
        decision,
        decisionRT,
        is_correct,
    )


def confidenceRatingTask(
    parameters: dict,
) -> Tuple[Optional[float], Optional[float], bool, Optional[float]]:
    """Confidence rating scale, using keyboard or mouse inputs.

    Parameters
    ----------
    parameters : dict
        Parameters dictionary.

    """

    from psychopy import core, visual

    print("...starting confidence rating.")

    # Initialise default values
    confidence, confidenceRT = None, None

    if parameters["device"] == "keyboard":
        markerStart = np.random.choice(
            np.arange(parameters["confScale"][0], parameters["confScale"][1])
        )
        ratingScale = visual.RatingScale(
            parameters["win"],
            low=parameters["confScale"][0],
            high=parameters["confScale"][1],
            noMouse=True,
            labels=parameters["labelsRating"],
            acceptKeys="down",
            markerStart=markerStart,
        )

        message = visual.TextStim(
            parameters["win"],
            height=parameters["textSize"],
            text=parameters["texts"]["Confidence"],
        )

        # Wait for response
        ratingProvided = False
        clock = core.Clock()
        while clock.getTime() < parameters["maxRatingTime"]:
            if not ratingScale.noResponse:
                ratingScale.markerColor = (0, 0, 1)
                if clock.getTime() > parameters["minRatingTime"]:
                    ratingProvided = True
                    break
            ratingScale.draw()
            message.draw()
            parameters["win"].flip()

        confidence = ratingScale.getRating()
        confidenceRT = ratingScale.getRT()

    elif parameters["device"] == "mouse":
        # Use the mouse position to update the slider position
        # The mouse movement is limited to a rectangle above the Slider
        # To avoid being dragged out of the screen (in case of multi screens)
        # and to avoid interferences with the Slider when clicking.
        parameters["win"].mouseVisible = False
        parameters["myMouse"].setPos((np.random.uniform(-0.25, 0.25), 0.2))
        parameters["myMouse"].clickReset()
        message = visual.TextStim(
            parameters["win"],
            height=parameters["textSize"],
            pos=(0, 0.2),
            text=parameters["texts"]["Confidence"],
        )
        slider = visual.Slider(
            win=parameters["win"],
            name="slider",
            pos=(0, -0.2),
            size=(0.7, 0.1),
            labels=parameters["texts"]["VASlabels"],
            granularity=1,
            ticks=(0, 100),
            style=("rating"),
            color="LightGray",
            flip=False,
            labelHeight=0.1 * 0.6,
        )
        slider.marker.size = (0.03, 0.03)
        clock = core.Clock()
        parameters["myMouse"].clickReset()
        buttons, confidenceRT = parameters["myMouse"].getPressed(getTime=True)

        while True:
            parameters["win"].mouseVisible = False
            trialdur = clock.getTime()
            buttons, confidenceRT = parameters["myMouse"].getPressed(getTime=True)

            # Mouse position (keep in in the rectangle)
            newPos = parameters["myMouse"].getPos()
            if newPos[0] < -0.5:
                newX = -0.5
            elif newPos[0] > 0.5:
                newX = 0.5
            else:
                newX = newPos[0]
            if newPos[1] < 0.1:
                newY = 0.1
            elif newPos[1] > 0.3:
                newY = 0.3
            else:
                newY = newPos[1]
            parameters["myMouse"].setPos((newX, newY))

            # Update marker position in Slider
            p = newX / 0.5
            slider.markerPos = 50 + (p * 50)

            # Check if response provided
            if (buttons == [1, 0, 0]) & (trialdur > parameters["minRatingTime"]):
                confidence, confidenceRT, ratingProvided = (
                    slider.markerPos,
                    clock.getTime(),
                    True,
                )
                print(
                    f"... Confidence level: {confidence}"
                    + f" with response time {round(confidenceRT, 2)} seconds"
                )
                # Change marker color after response provided
                slider.marker.color = "green"
                slider.draw()
                message.draw()
                parameters["win"].flip()
                core.wait(0.2)
                break
            elif trialdur > parameters["maxRatingTime"]:  # if too long
                ratingProvided = False
                confidenceRT = parameters["myMouse"].clickReset()

                # Text feedback if no rating provided
                message = visual.TextStim(
                    parameters["win"],
                    height=parameters["textSize"],
                    text="Too late",
                    color="red",
                    pos=(0.0, -0.2),
                )
                message.draw()
                parameters["win"].flip()
                core.wait(0.5)
                break
            slider.draw()
            message.draw()
            parameters["win"].flip()
    ratingEndTrigger = time.time()
    parameters["win"].flip()

    return confidence, confidenceRT, ratingProvided, ratingEndTrigger



# Author: Nicolas Legrand <nicolas.legrand@cas.au.dk>

from typing import List

import numpy as np
import pandas as pd


def cumulative_normal(x, alpha, beta):
    import pytensor.tensor as pt

    # Cumulative distribution function for the standard normal distribution
    return 0.5 + 0.5 * pt.erf((x - alpha) / (beta * pt.sqrt(2)))


def psychophysics(
    summary_df: pd.DataFrame,
    variables: List[str] = ["participant_id", "Modality"],
    additional_variables=[],
) -> pd.DataFrame:
    r"""Extract psychometric parameters from a set of result files from the HRD task.

    This function will use a Bayesian model to estimate psychophysics parameters and
    perform inference using MCMC sampling. The following parameters are returned:

    * Interoceptive bias

        * `bayesian_threshold` (the mean of the interoceptive bias)

        * `bayesian_slope` (the slope of the interoceptive bias)

    The interoceptive bias :math:`\alpha` represents the difference between the real
    heart rate and the cardiac belief. The interoceptive slope :math:`\beta` represents
    the precision of this bias (the standard deviation of the underlying cumulative
    normal function). These parameters are estimated using the following model:

    .. math::

        r_{i} & \sim \mathcal{Binomial}(\theta_{i},n_{i}) \\
        \Phi_{i}(x_{i}, \alpha, \beta) & = \frac{1}{2} + \frac{1}{2} * erf(\frac{x_{i}
        - \alpha}{\beta * \sqrt{2}}) \\
        \alpha & \sim \mathcal{Uniform}(-50.5, 50.5) \\
        \beta & \sim  \mathcal{Uniform}(.1, 30.0) \\

    Here :math:`x_i` is the proportion of positive response at the intensity :math:`i`.
    To compute the interoceptive bias, we use the `Alpha` value (the difference between
    the real heart rate and the tone that is presented at each trial). A negative value
    means that the tone needs to be slower than the heart rate for the participant to
    find it the same.

    * Cardiac beliefs

        * `belief_mean`

        * `belief_std`

    The mean of the cardiac belief :math:`\psi_{alpha}` represents the cardiac frequency
    that was inferred on average through the task. The precision of the cardiac belief
    :math:`\psi_{beta}` is the standard deviation around this belief. Under the
    hypothesis that the participant is not using any interoceptive information to
    perform the task, this value is the belief used to inform the decision by comparing
    it to the tones. These parameters are estimated using the following model:

    .. math::

        r_{i} & \sim \mathcal{Binomial}(\theta_{i},n_{i}) \\
        \Phi_{i}(x_{i}, \psi_{alpha}, \psi_{beta}) & = \frac{1}{2} + \frac{1}{2} *
        erf(\frac{x_{i} - \psi_{alpha}}{\psi_{beta} * \sqrt{2}}) \\
        \psi_{alpha} & \sim \mathcal{Uniform}(15.0, 200.0) \\
        \psi_{beta} & \sim  \mathcal{Uniform}(.1, 50.0) \\

    Here :math:`x_i` is the proportion of positive response at the intensity :math:`i`.
    To compute the interoceptive bias, we use the frequency of the tone presented
    during the decision phase only (assuming therefore that this is the only source of
    information used by the participant). The units are beat per minute (bpm).

    .. note::
        In the two equations above, $erf$ denotes the
        `error functions <https://en.wikipedia.org/wiki/Error_function>`_ and :math:`\phi`
        is the cumulative normal function.

    * Heart rate

        * `hr_mean` the mean of the averaged heart rates

        * `hr_std` the standard deviation of the averaged heart rates

    The mean of the averaged heart rates :math:`\omega_{alpha}` and the standard
    deviation of the averaged heart rates :math:`\omega_{beta}` are computed using the
    following model:

    .. math::

        r_{i} & \sim \mathcal{Normal}(\omega_{alpha},\omega_{beta}) \\
        \omega_{alpha} & \sim \mathcal{Uniform}(15.0, 200.0) \\
        \omega_{beta} & \sim  \mathcal{Uniform}(.1, 50.0) \\

    Here :math:`x_i` is the average heart rate at each trial.

    .. note::
        The heart rate that was recorded on every trial is the average of what was
        recorded over the 5 seconds of interoception during the listening phase. Here
        we are returning the mean and standard deviation of these values.

    .. warning::
        This function requires `PyMC <https://github.com/pymc-devs/pymc>`_.

    Parameters
    ----------
    summary_df :
        The data frame merges the individual result data frames. Multiple variables/
        condition can be specified using separate columns with the `variables` argument.
    variables :
        The variables coding for group/repeated measures. The default is
        `participant_id` and `Modality`.
    additional_variables :
        Additional variables for group/repeated measures.

    Returns
    -------
    results_df :
        The data frame containing, for each participant/condition/group, the
        psychometric variables.
    """
    import pymc as pm

    # create a list of variables to use to group the dataframe
    variables.extend(additional_variables)

    # the final data fram where results are saved
    results_df = pd.DataFrame()

    print("Extracting psychometric parameters from a large data frame.")
    print(f"... Independent variables provided: {variables}.")
    print(f"... {len(list(summary_df.groupby(variables)))} conditions in total.")

    # extract psychophysics parameters from trials for each sub data frame
    bias_x_total, bias_n_total, bias_r_total, bias_sub_total = [], [], [], []  # bias
    beliefs_x_total, beliefs_n_total, beliefs_r_total, beliefs_sub_total = (
        [],
        [],
        [],
        [],
    )  # beliefs
    hr_total, hr_sub_total = [], []  # heart rate
    print("... Extract trial-level psychophysics variables.")
    for i, grouped in enumerate(list(summary_df.groupby(variables))):
        cols, sub_df = grouped

        # update the independent variables
        results_df = pd.concat(
            [results_df, pd.Series(cols, index=variables).to_frame().T],
            ignore_index=True,
        )

        # extract trial-level psychometric parameters for bias
        # ------------------------------------------------------------------------------

        # intensity level, number of trials, number of positive responses
        x, n, r = np.zeros(203), np.zeros(203), np.zeros(203)
        for ii, intensity in enumerate(np.arange(-50.5, 51, 0.5)):
            x[ii] = intensity
            n[ii] = sum(sub_df.Alpha == intensity)
            r[ii] = sum((sub_df.Alpha == intensity) & (sub_df.Decision == "More"))

        # remove no responses trials
        validmask = n != 0
        xij, nij, rij = x[validmask], n[validmask], r[validmask]
        sub_vec = [i] * len(xij)

        bias_x_total.extend(xij)
        bias_n_total.extend(nij)
        bias_r_total.extend(rij)
        bias_sub_total.extend(sub_vec)

        # extract trial-level psychometric parameters for beliefs
        # ------------------------------------------------------------------------------

        # intensity level, number of trials, number of positive responses
        x, n, r = np.zeros(370), np.zeros(370), np.zeros(370)
        for ii, intensity in enumerate(np.arange(15, 200, 0.5)):
            x[ii] = intensity
            n[ii] = sum(sub_df.responseBPM == intensity)
            r[ii] = sum((sub_df.responseBPM == intensity) & (sub_df.Decision == "More"))

        # remove no responses trials
        validmask = n != 0
        xij, nij, rij = x[validmask], n[validmask], r[validmask]
        sub_vec = [i] * len(xij)

        beliefs_x_total.extend(xij)
        beliefs_n_total.extend(nij)
        beliefs_r_total.extend(rij)
        beliefs_sub_total.extend(sub_vec)

        # extract trial-level heart rate
        # ------------------------------------------------------------------------------

        # intensity level, number of trials, number of positive responses
        hr = sub_df.responseBPM.to_numpy()
        sub_vec = [i] * len(hr)

        hr_total.extend(hr)
        hr_sub_total.extend(sub_vec)

    # get the number of models to fit
    n = len(list(summary_df.groupby(variables)))

    # fit the model (thresholds and slopes)
    print("... Create the model and sample")
    with pm.Model():
        # Heart Rate -------------------------------------------------------------------
        hr_mean = pm.Uniform("hr_mean", lower=15.0, upper=200.0, shape=n)
        hr_std = pm.Uniform("hr_std", lower=0.1, upper=50.0, shape=n)
        _ = pm.Normal(
            "heart_rate",
            mu=hr_mean[hr_sub_total],
            sigma=hr_std[hr_sub_total],
            observed=hr_total,
        )

        # Cardiac beliefs --------------------------------------------------------------
        belief_mean = pm.Uniform("belief_mean", lower=15.0, upper=200.0, shape=n)
        belief_std = pm.Uniform("belief_std", lower=0.1, upper=50.0, shape=n)
        theta_beliefs = pm.Deterministic(
            "theta_beliefs",
            cumulative_normal(
                beliefs_x_total,
                belief_mean[beliefs_sub_total],
                belief_std[beliefs_sub_total],
            ),
        )
        _ = pm.Binomial(
            "p_beliefs", p=theta_beliefs, n=beliefs_n_total, observed=beliefs_r_total
        )

        # Slope and Threshold ----------------------------------------------------------
        threshold = pm.Uniform("threshold", lower=-50.5, upper=50.5, shape=n)
        slope = pm.Uniform("slope", lower=0.1, upper=30.0, shape=n)
        theta_bias = pm.Deterministic(
            "theta_bias",
            cumulative_normal(
                bias_x_total, threshold[bias_sub_total], slope[bias_sub_total]
            ),
        )
        _ = pm.Binomial("p_bias", p=theta_bias, n=bias_n_total, observed=bias_r_total)

        # sample
        idata = pm.sample(chains=4, cores=4)

    # save the mean of the parameter in the final dataframe
    results_df["bayesian_threshold"] = idata.posterior.threshold.mean(
        axis=(0, 1)
    ).to_numpy()
    results_df["bayesian_slope"] = idata.posterior.slope.mean(axis=(0, 1)).to_numpy()
    results_df["belief_mean"] = idata.posterior.belief_mean.mean(axis=(0, 1)).to_numpy()
    results_df["belief_std"] = idata.posterior.belief_std.mean(axis=(0, 1)).to_numpy()
    results_df["hr_mean"] = idata.posterior.hr_mean.mean(axis=(0, 1)).to_numpy()
    results_df["hr_std"] = idata.posterior.hr_std.mean(axis=(0, 1)).to_numpy()

    return results_df


def behaviours(
    summary_df: pd.DataFrame,
    variables: List[str] = ["participant_id", "Modality"],
    additional_variables=[],
) -> pd.DataFrame:
    r"""Extract behavioural parameters from a set of result files from the HRD task.

    For each participant/repeated measure/group, the following parameters are
    returned:

    * threshold
        The threshold of the psychometric curve as estimated during the task by the Psi
        staircase.
    * slope
        The slope of the psychometric curve as estimated during the task by the Psi
        staircase.
    * decision_mean_rt
        The average response time to decide whether the tone is faster or slower than
        the heart rate.
    * decision_median_rt
        The median response time to decide whether the tone is faster or slower than
        the heart rate.
    * confidence_mean_rt
        The average response time to provide the confidence ratings.
    * confidence_median_rt
        The median response time to provide the confidence ratings.
    * confidence_mean
        The average confidence level (using the same scale as what was used during
        the task).
    * dprime
        The sensitivity (SDT indices) in discriminating whether the tone is faster than
        the heart rate or not.
    * criterion
        The bias (SDT indices) in discriminating whether the tone is faster than the
        heart rate or not.

    .. warning::
        This function requires `metadpy <https://github.com/LegrandNico/metadpy>`_.

    Parameters
    ----------
    summary_df :
        The data frame merges the individual result data frames. Multiple variables /
        condition can be specified using separate columns with the `variables` argument.
    variables :
        The variables coding for group / repeated measures. The default is
        `participant_id` and `Modality`.
    additional_variables :
        Additional variables for group / repeated measures.

    Returns
    -------
    results_df :
        The data frame containing, for each participant/condition/group, the
        psychometric variables.
    """
    from metadpy import sdt

    # create a list of variables to use to group the dataframe
    variables.extend(additional_variables)

    # the final data fram where results are saved
    results_df = pd.DataFrame()

    print("Extracting behavioural indices from a large data frame.")
    print(f"... Independent variables provided: {variables}.")
    print(f"... {len(list(summary_df.groupby(variables)))} conditions in total.")

    for grouped in list(summary_df.groupby(variables)):
        cols, sub_df = grouped

        # psychophysics (Psi estimates)
        threshold = sub_df.EstimatedThreshold.dropna().iloc[-1]
        slope = sub_df.EstimatedSlope.dropna().iloc[-1]

        # response time
        # -------------
        decision_mean_rt = sub_df.DecisionRT.mean()
        decision_median_rt = sub_df.DecisionRT.median()

        confidence_mean_rt = sub_df.ConfidenceRT.mean()
        confidence_median_rt = sub_df.ConfidenceRT.median()

        # confidence
        # ----------
        confidence_mean = sub_df.Confidence.mean()

        # signal detection theory metrics
        # -------------------------------
        sub_df["Stimuli"] = sub_df.responseBPM > sub_df.listenBPM
        sub_df["Responses"] = sub_df.Decision == "More"

        # check that both signals have at least 5 valid trials each
        if (sub_df["Stimuli"].sum() > 5) & ((~sub_df["Stimuli"]).sum() > 5):
            hit, miss, fa, cr = sub_df.scores()
            hr, far = sdt.rates(hits=hit, misses=miss, fas=fa, crs=cr)
            dprime, criterion = sdt.dprime(hit_rate=hr, fa_rate=far), sdt.criterion(
                hit_rate=hr, fa_rate=far
            )
        else:
            (
                dprime,
                criterion,
            ) = (
                None,
                None,
            )

        # update the independent variables
        new_row = pd.Series(cols, index=variables).to_frame().T
        new_row["threshold"] = threshold
        new_row["slope"] = slope
        new_row["decision_mean_rt"] = decision_mean_rt
        new_row["decision_median_rt"] = decision_median_rt
        new_row["confidence_mean_rt"] = confidence_mean_rt
        new_row["confidence_median_rt"] = confidence_median_rt
        new_row["confidence_mean"] = confidence_mean
        new_row["dprime"] = dprime
        new_row["criterion"] = criterion

        results_df = pd.concat(
            [results_df, new_row],
            ignore_index=True,
        )

    return results_df


def metacognition(
    summary_df: pd.DataFrame,
    variables: List[str] = ["participant_id", "Modality"],
    additional_variables=[],
    bayesian: bool = True,
) -> pd.DataFrame:
    r"""Extract metacognitive parameters from a set of result files from the HRD task.

    For each participant/repeated measure/group, the following parameters are
    returned:


    .. warning::
        This function requires `metadpy <https://github.com/LegrandNico/metadpy>`_.

    Parameters
    ----------
    summary_df :
        The data frame merges the individual result data frames. Multiple variables/
        condition can be specified using separate columns with the `variables` argument.
    variables :
        The variables coding for group/repeated measures. The default is
        `participant_id` and `Modality`.
    additional_variables :
        Additional variables for group/repeated measures.

    Returns
    -------
    results_df :
        The data frame containing, for each participant/condition/group, the
        psychometric variables.
    """



# Author: Nicolas Legrand <nicolas.legrand@cas.au.dk>

import os
import subprocess
from os import PathLike
from pathlib import Path
from typing import List, Optional, Union

import numpy as np
import pandas as pd
import pkg_resources  # type: ignore

from cardioception.stats import behaviours, psychophysics


def cumulative_normal(x, alpha, beta):
    import pytensor.tensor as pt

    # Cumulative distribution function for the standard normal distribution
    return 0.5 + 0.5 * pt.erf((x - alpha) / (beta * pt.sqrt(2)))


def group_level_preprocessing(
    results: Union[PathLike, pd.DataFrame],
    variables: List[str] = ["participant_id", "Modality"],
    additional_variables=[],
    behavioural_indices: bool = True,
    psychophysical_indices: bool = True,
    metacognitive_indices: bool = True,
) -> pd.DataFrame:
    """Extrat all relevant indices from large result data frames.

    .. note::
        This function concatenate the results from
        {ref}`cardioception.stats.psychophysics`, {ref}`cardioception.stats.behaviours`
        and {ref}`cardioception.stats.metacognition`, see the documentation of thoses
        functions for more details on the indices.

    Parameters
    ----------
    results :
        The data frame merging the individual result data frames. Multiple variables /
        condition can be specifyed using separate columns with the `variables` argument.
    variables :
        The variables coding for group / repeated measures. The default is
        `participant_id` and `Modality`.
    additional_variables :
        Additional variables for group / repeated measures.
    behavioural_indices :
        Whether to extract the behavioural indices. Defaults to `True`.
    psychophysical_indices :
        Whether to extract the psychophysical indices. Defaults to `True`.
    metacognitive_indices
        Whether to extract the metacognitive indices. Defaults to `True`.

    Return
    ------

    See Also
    --------
    cardioception.stats.psychophysics, cardioception.stats.behaviours,
    cardioception.stats.metacognition

    """
    # read the input file if only the path was provided
    if not isinstance(results, pd.DataFrame):
        results_df = pd.read_csv(results)

    # create a list of variables to use to group the dataframe
    variables.extend(additional_variables)

    summary_df = pd.DataFrame([])

    if behavioural_indices:
        behaviours_df = behaviours(
            summary_df=results_df,
            variables=variables,
            additional_variables=additional_variables,
        )
        summary_df = pd.merge(left=summary_df, right=behaviours_df, on=variables)

    if psychophysical_indices:
        psychophysics_df = psychophysics(
            summary_df=results_df,
            variables=variables,
            additional_variables=additional_variables,
        )
        summary_df = pd.merge(left=summary_df, right=psychophysics_df, on=variables)

    if metacognitive_indices:
        pass

    return summary_df


def preprocessing(results: Union[PathLike, pd.DataFrame]) -> pd.DataFrame:
    """From the main behavioural data frame, extract summary metrics of behavioural,
    metacognitive and interoceptive performances.

    The slope and thresholds of the interoceptive/exteroceptive psychometric function
    are reported both using the online estimate outputted by the Psi staircase (i.e.
    `slope` and `threshold`), and using a Bayesian estimation (i.e. `bayesian_slope` and
    `bayesian_threshold`). The Bayesian estimation is the recommended value to use to
    report the results. Removing outliers before fitting will change the estimation,
    which is not the case for the Psi values.

    The d-prime and criterion are also computed using a classical SDT approach
    (`dprime` and `criterion`), as well as a Bayesian estimation performed when
    estimating the metacognitive sensitivity meta-d' (`bayesian_dprime`,
    `bayesian_criterion`, `bayesian_meta_d`, `bayesian_m_ratio`). The dprime and
    criterion can vary between the two methods. It is recommended to use the estimates
    consistently. Before the estimation of SDT and metacognitive metrics, the function
    ensure that at least 5 valid trials of each signal are present, otherwise returns
    `None`.

    When using this function for analysing results from the Heart Rate Discrimination
    task, the following packages should be credited: Systole [1]_, metadpy [2]_ and
    cardioception [3]_.

    Parameters
    ----------
    results : pd.DataFrame | PathLike
        Either the path to the result file, or the Pandas Data Frame.

    Returns
    -------
    summary_df : pd.DataFrame
        The summary statistic for this participant, splitting for interoception and
        exteroception if the two conditions were used.

    Notes
    -----
    This function will require [PyMC](https://github.com/pymc-devs/pymc) (>= 5.0) and
    [metadpy](https://github.com/LegrandNico/metadpy) (>=0.1.0).

    References
    ----------
    .. [1] Legrand et al., (2022). Systole: A python package for cardiac signal
        synchrony and analysis. Journal of Open Source Software, 7(69), 3832,
        https://doi.org/10.21105/joss.03832
    .. [2] https://github.com/LegrandNico/metadpy
    .. [3] Legrand, N., Nikolova, N., Correa, C., Brændholt, M., Stuckert, A., Kildahl,
        N., Vejlø, M., Fardo, F., & Allen, M. (2021). The Heart Rate Discrimination
        Task: A psychophysical method to estimate the accuracy and precision of
        interoceptive beliefs. Biological Psychology, 108239.
        https://doi.org/10.1016/j.biopsycho.2021.108239

    """
    import arviz as az
    import pymc as pm
    from metadpy import bayesian, sdt
    from metadpy.utils import discreteRatings

    # read the input file if only the path was provided
    if not isinstance(results, pd.DataFrame):
        results = pd.read_csv(results)

    summary_df = pd.DataFrame([])

    for modality in ["Intero", "Extero"]:
        this_modality = results[results.Modality == modality].copy()

        if len(this_modality) > 10:
            # response time
            # -------------
            decision_mean_rt = this_modality.DecisionRT.mean()
            decision_median_rt = this_modality.DecisionRT.median()

            confidence_mean_rt = this_modality.ConfidenceRT.mean()
            confidence_median_rt = this_modality.ConfidenceRT.median()

            # signal detection theory metrics
            # -------------------------------
            this_modality["Stimuli"] = (
                this_modality.responseBPM > this_modality.listenBPM
            )
            this_modality["Responses"] = this_modality.Decision == "More"

            # check that both signals have at least 5 valid trials each
            if (this_modality["Stimuli"].sum() > 5) & (
                (~this_modality["Stimuli"]).sum() > 5
            ):
                hit, miss, fa, cr = this_modality.scores()
                hr, far = sdt.rates(hits=hit, misses=miss, fas=fa, crs=cr)
                d, c = sdt.dprime(hit_rate=hr, fa_rate=far), sdt.criterion(
                    hit_rate=hr, fa_rate=far
                )
            else:
                (
                    d,
                    c,
                ) = (
                    None,
                    None,
                )

            # metacognitive sensitivity
            # -------------------------
            (
                bayesian_dprime,
                bayesian_criterion,
                bayesian_meta_d,
                bayesian_m_ratio,
            ) = (None, None, None, None)

            this_modality = this_modality[
                ~this_modality.Confidence.isna()
            ].copy()  # Drop trials with NaN in confidence rating
            this_modality.loc[:, "Accuracy"] = (
                (this_modality["Stimuli"] & this_modality["Responses"])
                | (~this_modality["Stimuli"] & ~this_modality["Responses"])
            ).copy()

            # check that both signals have at least 5 valid trials each
            if (this_modality["Stimuli"].sum() > 5) & (
                (~this_modality["Stimuli"]).sum() > 5
            ):
                try:
                    new_ratings, _ = discreteRatings(
                        this_modality.Confidence.to_numpy(), verbose=False
                    )
                    this_modality.loc[:, "discrete_confidence"] = new_ratings

                    metad = bayesian.hmetad(
                        data=this_modality,
                        stimuli="Stimuli",
                        accuracy="Accuracy",
                        confidence="discrete_confidence",
                        nRatings=4,
                        output="dataframe",
                    )
                    bayesian_dprime = metad["d"].values[0]
                    bayesian_criterion = metad["c"].values[0]
                    bayesian_meta_d = metad["meta_d"].values[0]
                    bayesian_m_ratio = metad["m_ratio"].values[0]

                except ValueError:
                    print(
                        (
                            f"Cannot discretize ratings for modality: {modality}. "
                            "The metacognitive efficiency will not be reported."
                        )
                    )

            # bayesian psychophysics
            # ----------------------
            x, n, r = np.zeros(203), np.zeros(203), np.zeros(203)

            for ii, intensity in enumerate(np.arange(-50.5, 51, 0.5)):
                x[ii] = intensity
                n[ii] = sum(this_modality.Alpha == intensity)
                r[ii] = sum(
                    (this_modality.Alpha == intensity)
                    & (this_modality.Decision == "More")
                )
            validmask = n != 0  # remove no responses trials
            xij, nij, rij = x[validmask], n[validmask], r[validmask]

            with pm.Model():
                alpha = pm.Uniform("alpha", lower=-40.5, upper=40.5)
                beta = pm.HalfNormal("beta", 10)
                thetaij = pm.Deterministic(
                    "thetaij", cumulative_normal(xij, alpha, beta)
                )
                _ = pm.Binomial("rij", p=thetaij, n=nij, observed=rij)
                idata = pm.sample(chains=4, cores=4)
            res = az.summary(idata)
            bayesian_threshold = res["mean"].alpha
            bayesian_slope = res["mean"].beta

            # Psi estimates
            threshold = this_modality.EstimatedThreshold.iloc[-1]
            slope = this_modality.EstimatedSlope.iloc[-1]

            # concatenate the summary statistics
            summary_df = pd.concat(
                [
                    summary_df,
                    pd.DataFrame(
                        {
                            "modality": modality,
                            "decision_mean_rt": decision_mean_rt,
                            "decision_median_rt": decision_median_rt,
                            "confidence_mean_rt": confidence_mean_rt,
                            "confidence_median_rt": confidence_median_rt,
                            "dprime": d,
                            "criterion": c,
                            "bayesian_dprime": bayesian_dprime,
                            "bayesian_criterion": bayesian_criterion,
                            "bayesian_meta_d": bayesian_meta_d,
                            "bayesian_m_ratio": bayesian_m_ratio,
                            "threshold": threshold,
                            "slope": slope,
                            "bayesian_threshold": bayesian_threshold,
                            "bayesian_slope": bayesian_slope,
                        },
                        index=[0],
                    ),
                ],
                ignore_index=True,
            )

    return summary_df


def report(
    result_path: PathLike, report_path: Optional[PathLike] = None, task: str = "HRD"
):
    """From the results folders, create HTML reports of behavioural and physiological
    data.

    Parameters
    ----------
    resultPath : PathLike
        Path variable. Where the results are stored (one participant only).
    reportPath : PathLike, optional
        Where the HTML report should be saved. If `None`, default will be in the
        provided `resultPath`.
    task : str, optional
        The task ("HRD" or "HBC"), by default "HRD".

    """
    from papermill import execute_notebook

    if report_path is None:
        report_path = result_path
    temp_notebook = Path(report_path, "temp.ipynb")
    htmlreport = Path(report_path, f"{task}_report.html")

    if task == "HRD":
        template = "HeartRateDiscrimination.ipynb"
    elif task == "HBC":
        template = "HeartBeatCounting.ipynb"

    execute_notebook(
        pkg_resources.resource_filename("cardioception.notebooks", template),
        temp_notebook,
        parameters=dict(resultPath=str(result_path), reportPath=str(report_path)),
    )
    command = (
        "jupyter nbconvert --to html --execute "
        + f"--TemplateExporter.exclude_input=True {temp_notebook} --output {htmlreport}"
    )
    subprocess.call(command, shell=True)
    os.remove(temp_notebook)



# Author: Nicolas Legrand <nicolas.legrand@cas.au.dk>

import os
import unittest
from unittest import TestCase
import pandas as pd
from pathlib import Path
from cardioception.reports import report, preprocessing


class TestReports(TestCase):
    def test_preprocessing(self):
        """Test the preprocessing function"""
        # load the main result data frame
        results = pd.read_csv(
            "https://raw.githubusercontent.com/LegrandNico/Cardioception/master/docs/source/examples/templates/data/HRD/HRD_final.txt"
            )
        preprocessing(results=results)


    def test_report(self):
        """Test the report function"""

        #####
        # HRD
        #####
        hrd_results_path = Path(Path.cwd(), "docs", "source", "examples", "templates", "data", "HRD")
        hrd_report_path = Path.cwd()
        report(result_path=hrd_results_path, report_path=hrd_report_path)
        #os.remove(Path(hrd_report_path, "HRD_report.html"))

        #####
        # HBC
        #####
        hbc_results_path = Path(Path.cwd(), "docs", "source", "examples", "templates", "data", "HBC")
        hbc_report_path = Path.cwd()
        report(result_path=hbc_results_path, report_path=hbc_report_path, task="HBC")
        #os.remove(Path(hbc_report_path, "HBC_report.html"))


if __name__ == "__main__":
    unittest.main(argv=["first-arg-is-ignored"], exit=False)



# Author: Nicolas Legrand <nicolas.legrand@cas.au.dk>

from typing import Optional, Tuple

import numpy as np
import pandas as pd


def run(
    parameters: dict,
    runTutorial: bool = True,
):
    """Run the entire task sequence.

    Parameters
    ----------
    parameters : dict
        Task parameters.
    tutorial : bool
        If `True`, will present a tutorial with 10 training trial with feedback and 5
        trials with confidence rating.

    """

    from psychopy import core, visual

    # Run tutorial
    if runTutorial is True:
        tutorial(parameters)

    # Rest
    if parameters["restPeriod"] is True:
        rest(parameters, duration=parameters["restLength"])

    for condition, duration, nTrial in zip(
        parameters["conditions"],
        parameters["times"],
        range(0, len(parameters["conditions"])),
    ):
        parameters["triggers"]["trialStart"]  # Send trigger or None

        nCount, confidence, confidenceRT = trial(
            condition, duration, nTrial, parameters
        )

        parameters["triggers"]["trialStop"]  # Send trigger or None

        # Store results in a DataFrame
        parameters["results_df"] = pd.concat(
            [
                parameters["results_df"],
                pd.DataFrame(
                    {
                        "nTrial": [nTrial],
                        "Reported": [nCount],
                        "Condition": [condition],
                        "Duration": [duration],
                        "Confidence": [confidence],
                        "ConfidenceRT": [confidenceRT],
                    }
                ),
            ],
            ignore_index=True,
        )

        # Save the results at each iteration
        parameters["results_df"].to_csv(
            parameters["resultPath"]
            + "/"
            + parameters["participant"]
            + parameters["session"]
            + ".txt",
            index=False,
        )

    # Save results
    parameters["results_df"].to_csv(
        parameters["resultPath"]
        + "/"
        + parameters["participant"]
        + parameters["session"]
        + "_final.txt",
        index=False,
    )

    # End of the task
    end = visual.TextStim(
        parameters["win"],
        height=parameters["textSize"],
        pos=(0.0, 0.0),
        text="You have completed the task. Thank you for your participation.",
    )
    end.draw()
    parameters["win"].flip()
    core.wait(3)


def trial(
    condition: str,
    duration: int,
    nTrial: int,
    parameters: dict,
) -> Tuple[Optional[int], Optional[float], Optional[float]]:
    """Run one trial.

    Parameters
    ----------
    condition : str
        The trial condition, can be `"Rest"` or `"Count"`.
    duration : int
        The lenght of the recording (in seconds).
    ntrial : int
        Trial number.
    parameters : dict
        Task parameters.

    Returns
    -------
    nCount : int
        The number of heartbeat estimated by the participant.
    confidence : int
        The confidence in the estimation of the heartbeat provided by the
        participant.
    confidenceRT : float
        The response time to provide confidence rating.

    """

    from psychopy import core, event, visual

    # Initialize default values
    confidence, confidenceRT = None, None
    nCounts: str = ""

    # Ask the participant to press 'Space' (default) to start the trial
    messageStart = visual.TextStim(
        parameters["win"], height=parameters["textSize"], text="Press space to continue"
    )
    messageStart.draw()
    parameters["win"].flip()
    event.waitKeys(keyList=parameters["startKey"])
    parameters["win"].flip()

    parameters["oxiTask"].setup()
    parameters["oxiTask"].read(duration=2)

    # Show instructions
    if condition == "Rest":
        message = visual.TextStim(
            parameters["win"],
            text=parameters["texts"]["Rest"],
            pos=(0.0, 0.2),
            height=parameters["textSize"],
        )
        message.draw()
        parameters["restLogo"].draw()
    elif (condition == "Count") | (condition == "Training"):
        message = visual.TextStim(
            parameters["win"],
            text=parameters["texts"]["Count"],
            pos=(0.0, 0.2),
            height=parameters["textSize"],
        )
        message.draw()
        parameters["heartLogo"].draw()
    parameters["win"].flip()

    # Wait for a beat to start the task
    parameters["oxiTask"].waitBeat()
    core.wait(3)

    # Sound signaling trial start
    if (condition == "Count") | (condition == "Training"):
        parameters["oxiTask"].readInWaiting()
        # Add event marker
        parameters["oxiTask"].channels["Channel_0"][-1] = 1
        parameters["noteStart"].play()
        parameters["triggers"]["listeningStart"]
        core.wait(1)

    # Record for a desired time length
    parameters["oxiTask"].read(duration=duration - 1)

    # Sound signaling trial stop
    if (condition == "Count") | (condition == "Training"):
        # Add event marker
        parameters["oxiTask"].readInWaiting()
        parameters["oxiTask"].channels["Channel_0"][-1] = 2
        parameters["noteStop"].play()
        parameters["triggers"]["listeningStop"]
        core.wait(3)
        parameters["oxiTask"].readInWaiting()

    # Hide instructions
    parameters["win"].flip()

    # Save recording
    parameters["oxiTask"].save(
        parameters["resultPath"]
        + "/"
        + parameters["participant"]
        + str(nTrial)
        + "_"
        + str(nTrial)
    )

    ###############################
    # Record participant estimation
    ###############################
    if (condition == "Count") | (condition == "Training"):
        # Ask the participant to press 'Space' (default) to start the trial
        messageCount = visual.TextStim(
            parameters["win"],
            height=parameters["textSize"],
            pos=(0, 0.2),
            text=parameters["texts"]["nCount"],
        )
        messageCount.draw()
        parameters["win"].flip()

        parameters["triggers"]["decisionStart"]  # Send trigger or None

        nCounts = ""
        while True:
            # Record new key
            key = event.waitKeys(
                keyList=[
                    "escape",
                    "backspace",
                    "return",
                    "1",
                    "2",
                    "3",
                    "4",
                    "5",
                    "6",
                    "7",
                    "8",
                    "9",
                    "0",
                    "num_1",
                    "num_2",
                    "num_3",
                    "num_4",
                    "num_5",
                    "num_6",
                    "num_7",
                    "num_8",
                    "num_9",
                    "num_0",
                ]
            )

            if key[0] == "escape":
                keys = event.getKeys()
                if "escape" in keys:
                    print("User abort")
                    parameters["win"].close()
                    core.quit()
            if key[0] == "backspace":
                if nCounts:
                    nCounts = nCounts[:-1]
            elif key[0] == "return":
                if not all(char.isdigit() for char in nCounts):
                    messageError = visual.TextStim(
                        parameters["win"],
                        height=parameters["textSize"],
                        pos=(0, 0.2),
                        text="You should only provide numbers",
                    )
                    messageError.draw()
                    parameters["win"].flip()
                    core.wait(2)
                elif nCounts == "":
                    messageError = visual.TextStim(
                        parameters["win"],
                        height=parameters["textSize"],
                        pos=(0, 0.2),
                        text="You should provide numbers",
                    )
                    messageError.draw()
                    parameters["win"].flip()
                    core.wait(2)
                else:
                    break

            else:
                if key:
                    nCounts += [s for s in key[0] if s.isdigit()][0]

            # Show the text on the screen
            recordedText = visual.TextStim(
                parameters["win"], height=parameters["textSize"], text=nCounts
            )
            recordedText.draw()
            messageCount.draw()
            parameters["win"].flip()

        parameters["triggers"]["decisionStop"]  # Send trigger or None

        ##############
        # Rating scale
        ##############
        if parameters["rating"] is True:
            markerStart = np.random.choice(
                np.arange(parameters["confScale"][0], parameters["confScale"][1])
            )
            ratingScale = visual.RatingScale(
                parameters["win"],
                low=parameters["confScale"][0],
                high=parameters["confScale"][1],
                noMouse=True,
                labels=parameters["labelsRating"],
                acceptKeys="down",
                markerStart=markerStart,
            )
            message = visual.TextStim(
                parameters["win"],
                text=parameters["texts"]["confidence"],
                height=parameters["textSize"],
            )
            parameters["triggers"]["confidenceStart"]
            while ratingScale.noResponse:
                message.draw()
                ratingScale.draw()
                parameters["win"].flip()
            confidence = ratingScale.getRating()
            confidenceRT = ratingScale.getRT()
            parameters["triggers"]["confidenceStop"]

    finalCount = int(nCounts) if nCounts else None

    return finalCount, confidence, confidenceRT


def tutorial(parameters: dict):
    """Run tutorial for the Heartbeat Counting Task.

    Parameters
    ----------
    parameters : dict
        Task parameters.
    win : `psychopy.visual.window` or None
        The window in which to draw objects.
    """

    from psychopy import event, visual

    # Tutorial 1
    messageStart = visual.TextStim(
        parameters["win"],
        height=parameters["textSize"],
        text=parameters["texts"]["Tutorial1"],
    )
    messageStart.draw()
    press = visual.TextStim(
        parameters["win"],
        height=parameters["textSize"],
        text="Please press SPACE to continue",
        pos=(0.0, -0.4),
    )
    press.draw()
    parameters["win"].flip()
    event.waitKeys(keyList=parameters["startKey"])

    # Tutorial 2
    messageStart = visual.TextStim(
        parameters["win"],
        height=parameters["textSize"],
        pos=(0.0, 0.2),
        text=parameters["texts"]["Tutorial2"],
    )
    messageStart.draw()
    parameters["heartLogo"].draw()
    press = visual.TextStim(
        parameters["win"],
        height=parameters["textSize"],
        text="Please press SPACE to continue",
        pos=(0.0, -0.4),
    )
    press.draw()
    parameters["win"].flip()
    event.waitKeys(keyList=parameters["startKey"])

    # Tutorial 3
    if parameters["taskVersion"] == "Shandry":
        messageStart = visual.TextStim(
            parameters["win"],
            height=parameters["textSize"],
            pos=(0.0, 0.2),
            text=parameters["texts"]["Tutorial3"],
        )
        messageStart.draw()
        parameters["restLogo"].draw()
        press = visual.TextStim(
            parameters["win"],
            height=parameters["textSize"],
            text="Please press SPACE to continue",
            pos=(0.0, -0.4),
        )
        press.draw()
        parameters["win"].flip()
        event.waitKeys(keyList=parameters["startKey"])

    # Tutorial 4
    messageStart = visual.TextStim(
        parameters["win"],
        height=parameters["textSize"],
        text=parameters["texts"]["Tutorial4"],
    )
    messageStart.draw()
    press = visual.TextStim(
        parameters["win"],
        height=parameters["textSize"],
        text="Please press SPACE to continue",
        pos=(0.0, -0.4),
    )
    press.draw()
    parameters["win"].flip()

    event.waitKeys(keyList=parameters["startKey"])

    # Tutorial 5
    messageStart = visual.TextStim(
        parameters["win"],
        height=parameters["textSize"],
        text=parameters["texts"]["Tutorial5"],
    )
    messageStart.draw()
    press = visual.TextStim(
        parameters["win"],
        height=parameters["textSize"],
        text="Please press SPACE to continue",
        pos=(0.0, -0.4),
    )
    press.draw()
    parameters["win"].flip()
    event.waitKeys(keyList=parameters["startKey"])

    # Tutorial 6
    messageStart = visual.TextStim(
        parameters["win"],
        height=parameters["textSize"],
        text=parameters["texts"]["Tutorial6"],
    )
    messageStart.draw()
    press = visual.TextStim(
        parameters["win"],
        height=parameters["textSize"],
        text="Please press SPACE to continue",
        pos=(0.0, -0.4),
    )
    press.draw()
    parameters["win"].flip()
    event.waitKeys(keyList=parameters["startKey"])

    # Tutorial 7
    messageStart = visual.TextStim(
        parameters["win"],
        height=parameters["textSize"],
        text=parameters["texts"]["Tutorial7"],
    )
    messageStart.draw()
    press = visual.TextStim(
        parameters["win"],
        height=parameters["textSize"],
        text="Please press SPACE to continue",
        pos=(0.0, -0.4),
    )
    press.draw()
    parameters["win"].flip()
    event.waitKeys(keyList=parameters["startKey"])

    # Tutorial 8
    messageStart = visual.TextStim(
        parameters["win"],
        height=parameters["textSize"],
        text=parameters["texts"]["Tutorial8"],
    )
    messageStart.draw()
    press = visual.TextStim(
        parameters["win"],
        height=parameters["textSize"],
        text="Please press SPACE to continue",
        pos=(0.0, -0.4),
    )
    press.draw()
    parameters["win"].flip()
    event.waitKeys(keyList=parameters["startKey"])

    # Practice trial
    _ = trial("Count", 15, 0, parameters)

    # Tutorial 9
    messageStart = visual.TextStim(
        parameters["win"],
        height=parameters["textSize"],
        text=parameters["texts"]["Tutorial9"],
    )
    messageStart.draw()
    press = visual.TextStim(
        parameters["win"],
        height=parameters["textSize"],
        text="Please press SPACE to continue",
        pos=(0.0, -0.4),
    )
    press.draw()
    parameters["win"].flip()
    event.waitKeys(keyList=parameters["startKey"])


def rest(parameters: dict, duration: float = 300.0):
    """Run a resting state period for heart rate variability before running the Heart
    Beat Counting Task.

    Parameters
    ----------
    parameters : dict
        Task parameters.
    duration : float
        Duration or the recording (seconds).

    """

    from psychopy import visual

    # Show the resting state instructions
    messageStart = visual.TextStim(
        parameters["win"],
        height=parameters["textSize"],
        pos=(0.0, 0.2),
        text=("Calibrating... Please sit quietly" " until the end of the recording."),
    )
    messageStart.draw()
    parameters["restLogo"].draw()
    parameters["win"].flip()

    # Record PPG signal
    parameters["oxiTask"].setup()
    parameters["oxiTask"].read(duration=duration)

    # Save recording
    parameters["oxiTask"].save(
        parameters["resultPath"] + "/" + parameters["participant"] + "_Rest"
    )



# Author: Nicolas Legrand <nicolas.legrand@cas.au.dk>
from typing import Collection, Dict


def english(device: str, setup: str, exteroception: bool) -> Dict[str, Collection[str]]:
    """Create the text dictionary with instruction in Danish

    Parameters
    ----------
    device : str
        Can be `"keyboard"` or `"mouse"`.
    setup : str
        The experimental setup. Can be `"behavioral"` or `"test"`.
    exteroception : bool
        If `True`, the task includes and exteroceptive control condition.

    Returns
    -------
    texts : dict

    """
    btnext = "press SPACE" if device == "keyboard" else "click the mouse"
    texts = {
        "done": "You have completed the task. Thank you for your participation.",
        "slower": "Slower",
        "faster": "Faster",
        "checkOximeter": "Please make sure the oximeter is correctly clipped to your finger.",
        "stayStill": "Please stay still during the recording",
        "tooLate": "Too late",
        "correctResponse": "Correct",
        "incorrectResponse": "False",
        "VASlabels": ["Guess", "Certain"],
        "textHeartListening": "Listen to your heart",
        "textToneListening": "Listen to the tones",
        "textTaskStart": "The task is now going to start, get ready.",
        "textBreaks": f"Break. You can rest as long as you want. Just {btnext} when you want to resume the task.",
        "textNext": f"Please {btnext} to continue",
        "textWaitTrigger": "Waiting for fMRI trigger...",
        "Decision": {
            "Intero": """Are these beeps faster or slower than your heart?""",
            "Extero": """Are these beeps faster or slower than the previous?""",
        },
        "Confidence": """How confident are you in your choice?""",
    }

    if device == "keyboard":
        texts["responseText"] = "Use DOWN key for slower - UP key for faster."
    elif device == "mouse":
        texts["responseText"] = "Use LEFT button for slower - RIGHT button for faster."

    texts[
        "Tutorial1"
    ] = """During this experiment, we will record your pulse and play beeps based on your heart rate.

You will only be allowed to focus on the internal sensations of your heartbeats, but not to measure your heart rate by any other means (e.g. checking pulse at your wrist or your neck).
        """
    texts[
        "pulseTutorial1"
    ] = "Please place the pulse oximeter on your forefinger. Use your non-dominant hand as depicted in this schema."

    texts[
        "pulseTutorial2"
    ] = "If you can feel your heartbeats when you have the pulse oximeter on your forefinger, try to place it on another finger."

    texts[
        "pulseTutorial3"
    ] = "You can test different configurations until you find the finger which provides you with the least sensory input about your heart rate."

    texts[
        "pulseTutorial4"
    ] = "Please enter the number of the finger corresponding to the finger where you decided to place the pulse oximeter."

    texts[
        "Tutorial2"
    ] = "When you see this icon, try to focus on your heartbeat for 5 seconds. Try not to move, as we are recording your pulse in this period"

    moreResp = "UP key" if device == "keyboard" else "RIGHT mouse button"
    lessResp = "DOWN key" if device == "keyboard" else "LEFT mouse button"
    texts[
        "Tutorial3_icon"
    ] = """After this 'heart listening' period, you will see the same icon and hear a series of beeps."""
    texts[
        "Tutorial3_responses"
    ] = f"""As quickly and accurately as possible, you will listen to these beeps and decide if they are faster ({moreResp}) or slower ({lessResp}) than your own heart rate.

The beeps will ALWAYS be slower or faster than your heart. Please guess, even if you are unsure."""

    if exteroception is True:
        texts[
            "Tutorial3bis"
        ] = """For some trials, instead of seeing the heart icon, you will see a listening icon. You will then have to listen to a first set of beeps, instead of your heart."""

        texts[
            "Tutorial3ter"
        ] = f"""After these first beeps, you will see the response icons appear, and a second set of beeps will play.

As quickly and accurately as possible, you will listen to these beeps and decide if they are faster ({moreResp}) or slower ({lessResp}) than the first set of beeps.

The second series of beeps will ALWAYS be slower or faster than the first series. Please guess, even if you are unsure."""

    texts[
        "Tutorial4"
    ] = """Once you have provided your decision, you will also be asked to rate how confident you feel in your decision.

Here, the maximum rating means that you are totally certain in your choice, and the smallest rating means that you felt that you were guessing.

You should use mouse to select your rating"""

    texts[
        "Tutorial5"
    ] = """This sequence will be repeated during the task.

At times the task may be very difficult; the difference between your true heart rate and the presented beeps may be very small.

This means that you should try to use the entire length of the confidence scale to reflect your subjective uncertainty on each trial.

As the task difficulty will change over time, it is rare that you will be totally confident or totally uncertain."""

    texts[
        "Tutorial6"
    ] = """This concludes the tutorial. If you have any questions, please ask the experimenter now.
Otherwise, you can continue to the main task."""

    return texts


def danish(device: str, setup: str, exteroception: bool) -> Dict[str, Collection[str]]:
    """Create the text dictionary with instruction in Danish

    Parameters
    ----------
    device : str
        Can be `"keyboard"` or `"mouse"`.
    setup : str
        The experimental setup. Can be `"behavioral"` or `"test"`.
    exteroception : bool
        If `True`, the task includes and exteroceptive control condition.

    Returns
    -------
    texts : dict

    """

    btnext = "tryk på mellemrumstasten" if device == "keyboard" else "klik på musen"
    texts = {
        "done": "Du har genemført opgaven. Tak for din deltagalse.",
        "slower": "Langsommere",
        "faster": "Hurtigere",
        "checkOximeter": "Sørg venligst for at pulsoximeteret sidder rigtigt på din finger.",
        "stayStill": "Sid venligst roligt under målingen",
        "tooLate": "For langsomt",
        "correctResponse": "Rigtigt",
        "incorrectResponse": "Forkert",
        "VASlabels": ["Gæt", "Helt sikker"],
        "textHeartListening": "Mærk din hjerterytme",
        "textToneListening": "Lyt til tonerne",
        "textTaskStart": "Opgaven begynder nu, gør dig klar.",
        "textBreaks": f"Pause. Du kan tage så lang en pause, som du har brug for. Bare {btnext} når du vil fortsætte opgaven.",
        "textNext": f"Venligst, {btnext} for at fortsætte",
        "textWaitTrigger": "Venter på fMRI-udløseren...",
        "Decision": {
            "Intero": """Er disse bib-lyde hurtigere eller langsommere end dit hjerte?""",
            "Extero": """Er disse bib-lyde hurtigere eller langsommere end den de forrige? """,
        },
        "Confidence": """Hvor sikker er du på dit svar?""",
    }

    if device == "keyboard":
        texts[
            "responseText"
        ] = "Brug NED tasten for langsommere - OP tasten for hurtigere."
    elif device == "mouse":
        texts[
            "responseText"
        ] = "Brug VENSTRE museknap for langsommere - HØJRE museknap for hurtigere."

    texts[
        "Tutorial1"
    ] = """I dette forsøg vil vi registrere din puls og afspille bib-lyde baseret på din hjerterytme.

Du må kun fokusere på din indre følelse af din hjerterytme. Du må altså ikke måle din hjerterytme på andre måder (fx ved at tjekke din puls på dit håndled eller din hals).
        """
    texts[
        "pulseTutorial1"
    ] = "Placer venligst puls oximeteret på din pegefinger. Brug din ikke-dominante hånd som beskrevet i dette skema."

    texts[
        "pulseTutorial2"
    ] = "Hvis du kan mærke din hjerterytme, når du har puls oximeteret på din pegefinger, så prøv at placere det på en anden finger."

    texts[
        "pulseTutorial3"
    ] = "Du kan teste forskellige fingre indtil du finder den finger, der giver dig mindst sensorisk indput omkring din hjerterytme."

    texts[
        "pulseTutorial4"
    ] = "Indtast venligt nummeret på den finger som du besluttede at placere puls oximeteret på."

    texts[
        "Tutorial2"
    ] = "Når du ser dette ikon, forsøg da at fokusere på din hjerterytme i 5 sekunder. Prøv ikke at bevæge dig, da vi registrere din puls i dette tidsrum"

    moreResp = "OP tasten" if device == "keyboard" else "HØJRE mussetast"
    lessResp = "NED tasten" if device == "keyboard" else "VENSTRE mussetast"
    texts[
        "Tutorial3_icon"
    ] = """Efter tidsrummet hvor du har forsøgt at mærke dit hjerte, vil du se det samme ikon og høre en række bib-lyde."""
    texts[
        "Tutorial3_responses"
    ] = f"""Det følgende skal du gøre så hurtigt og præcist som muligt: Du vil lytte til disse bib-lyde og beslutte om de er hurtigere ({moreResp}) eller langsommere ({lessResp}) end din egen hjerterytme.

Bib-lydene vil ALTID være langsommere eller hurtigere end dit hjerte. Gæt venligst selvom du er usikker."""

    if exteroception is True:
        texts[
            "Tutorial3bis"
        ] = """I nogle runder vil du se et lytteikon i stedet for et hjerteikon. Her vil du skulle lytte til et sæt af bib-lyde i stedet for dit hjerte."""

        texts[
            "Tutorial3ter"
        ] = f"""Efter dette sæt af bib-lyde vil du se, at svarikonet dukker op, og et andet sæt af bib-lyde vil blive afspillet.

Det følgende skal du gøre så hurtigt og præcist som muligt: Du vil lytte til det sidste sæt af bib-lyde og beslutte om de er hurtigere ({moreResp}) eller langsommere ({lessResp}) end det første sæt af bib-lyde.

Det andet sæt af bib-lyde vil ALTID være langsommere eller hurtigere end det første sæt. Gæt venligst selvom du er usikker."""

    texts[
        "Tutorial4"
    ] = """Når du har svaret, vil du også blive bedt om at angive hvor sikker du er på din beslutning.

Her betyder den højeste score at du er helt sikker på dit valg, og den mindste score betyder, at du følte, at du gættede.

Du skal bruge musen til at vælge en score."""

    texts[
        "Tutorial5"
    ] = """Denne sekvens vil blive gentaget igennem opgaven.

Nogle gange kan opgaven være virkelig svær; forskellen mellem din faktiske hjerterytme og bib-lydene kan være meget små.

Dette betyder, at du skal forsøge at bruge hele skalaens længde til at angive din subjektive usikkerhed i hver runde.

Da opgavens sværhedsgrad ændrer sig over tid, er det sjældent at du vil være totalt sikker eller totalt usikker."""

    texts[
        "Tutorial6"
    ] = """Dette er slutningen på vejledningen. Hvis du har noget spørgsmål, så spørg endelig en forsker nu.
Ellers kan du fortsætte til hovedopgaven."""

    return texts


def danish_children(
    device: str, setup: str, exteroception: bool
) -> Dict[str, Collection[str]]:
    """Create the text dictionary with instruction in Danish (simplified version for
    children).

    Parameters
    ----------
    device : str
        Can be `"keyboard"` or `"mouse"`.
    setup : str
        The experimental setup. Can be `"behavioral"` or `"test"`.
    exteroception : bool
        If `True`, the task includes and exteroceptive control condition.

    Returns
    -------
    texts : dict

    """

    btnext = "tryk på mellemrumstasten" if device == "keyboard" else "klik på musen"
    texts = {
        "done": "Du har genemført opgaven. Tak for din deltagalse.",
        "slower": "Langsommere",
        "faster": "Hurtigere",
        "checkOximeter": "Spørg forskningsassistensen om, hvordan du skal placere fingerklemmen.",
        "stayStill": "Sid venligst roligt under målingen",
        "tooLate": "For langsomt",
        "correctResponse": "Rigtigt",
        "incorrectResponse": "Forkert",
        "VASlabels": ["Slet ikke sikker", "Helt sikker"],
        "textHeartListening": "Mærk din indre puls",
        "textToneListening": "Lyt til tonerne",
        "textTaskStart": "Opgaven begynder nu, gør dig klar.",
        "textBreaks": f"Pause. Du kan tage så lang en pause, som du har brug for. Bare {btnext} når du vil fortsætte opgaven.",
        "textNext": f"Venligst, {btnext} for at fortsætte",
        "textWaitTrigger": "Venter på fMRI-udløseren...",
        "Decision": {
            "Intero": """Er disse bib-lyde hurtigere eller langsommere end dit hjerte?""",
            "Extero": """Er disse bib-lyde hurtigere eller langsommere end den de forrige? """,
        },
        "Confidence": """Hvor sikker er du på dit svar?""",
    }

    if device == "keyboard":
        texts[
            "responseText"
        ] = "Brug NED tasten for langsommere - OP tasten for hurtigere."
    elif device == "mouse":
        texts[
            "responseText"
        ] = "Brug VENSTRE museknap for langsommere - HØJRE museknap for hurtigere."

    texts[
        "Tutorial1"
    ] = """Instruktion 1
        """
    texts["pulseTutorial1"] = "Udstyr."

    texts["pulseTutorial2"] = ""

    texts["pulseTutorial3"] = ""

    texts[
        "pulseTutorial4"
    ] = "Indtast venligt nummeret på den finger som du besluttede at placere fingerklemmen på."

    texts[
        "Tutorial2"
    ] = "Når du ser dette ikon, forsøg da at fokusere på din indre puls i 5 sekunder. Prøv ikke at bevæge dig, da vi måler din puls i dette tidsrum"

    moreResp = "OP tasten" if device == "keyboard" else "HØJRE mussetast"
    lessResp = "NED tasten" if device == "keyboard" else "VENSTRE mussetast"
    texts[
        "Tutorial3_icon"
    ] = """Efter du har forsøgt at mærke din indre puls, vil du se det samme ikon og høre en række bib-lyde."""
    texts["Tutorial3_responses"] = """Instruktion 2"""

    if exteroception is True:
        texts[
            "Tutorial3bis"
        ] = """I nogle runder vil du se et lytteikon i stedet for et hjerteikon. Her vil du skulle lytte til et sæt af bib-lyde i stedet for dit hjerte."""

        texts[
            "Tutorial3ter"
        ] = f"""Efter dette sæt af bib-lyde vil du se, at svarikonet dukker op, og et andet sæt af bib-lyde vil blive afspillet.

Det følgende skal du gøre så hurtigt og præcist som muligt: Du vil lytte til det sidste sæt af bib-lyde og beslutte om de er hurtigere ({moreResp}) eller langsommere ({lessResp}) end det første sæt af bib-lyde.

Det andet sæt af bib-lyde vil ALTID være langsommere eller hurtigere end det første sæt. Gæt venligst selvom du er usikker."""

    texts["Tutorial4"] = """Instruktion 3"""

    texts["Tutorial5"] = """Instruktion 4"""

    texts[
        "Tutorial6"
    ] = """Dette er slutningen på vejledningen. Hvis du har noget spørgsmål, så spørg endelig en forsker nu.
Ellers kan du fortsætte til opgaven."""

    return texts


def french(device: str, setup: str, exteroception: bool) -> Dict[str, Collection[str]]:
    """Create the text dictionary with instruction in french

    Parameters
    ----------
    device : str
        Can be `"keyboard"` or `"mouse"`.
    setup : str
        The experimental setup. Can be `"behavioral"` or `"test"`.
    exteroception : bool
        If `True`, the task includes and exteroceptive control condition.

    Returns
    -------
    texts : dict

    """
    btnext = (
        "appuyez sur la barre espace"
        if device == "keyboard"
        else "cliquez sur la souris"
    )
    texts = {
        "done": "Vous avez terminé la tâche. Merci pour votre participation.",
        "slower": "Plus lent",
        "faster": "Plus rapide",
        "checkOximeter": "Assurez-vous que l'oxymètre est bien attaché à votre doigt.",
        "stayStill": "Veuillez ne pas bouger pendant l'enregistrement",
        "tooLate": "Trop tard",
        "correctResponse": "Correct",
        "incorrectResponse": "Faux",
        "VASlabels": ["Incertain", "Tout à fait sûr"],
        "textHeartListening": "Ecoutez votre coeur",
        "textToneListening": "Ecoutez les sons",
        "textTaskStart": "La tâche va débuter, tenez-vous prêt.",
        "textBreaks": f"Pause. Vous pouvez vous reposer aussi longtemps que vous le souhaitez. Simplement {btnext} quand vous désirez rependre la tâche.",
        "textNext": f"S'il vous plaît {btnext} pour continuer",
        "textWaitTrigger": "Attendez pour le déclencheur IRMf...",
        "Decision": {
            "Intero": """Est-ce que ces sons sont plus rapides ou plus lents que votre coeur?""",
            "Extero": """Est-ce que ces sons sont plus rapides ou plus lents que les précédents?""",
        },
        "Confidence": """Etes-vous sûr de votre choix?""",
    }

    if device == "keyboard":
        texts[
            "responseText"
        ] = "Appuyez sur la flèche vers le BAS pour plus lent - vers le HAUT pour plus rapide."
    elif device == "mouse":
        texts[
            "responseText"
        ] = "Appuyez sur le clic GAUCHE pour plus lent - clic DROIT pour plus rapide."

    texts[
        "Tutorial1"
    ] = """Durant cette tâche, nous allons enregistrer vos pulsations et jouer des sons basés sur votre rythme cardiaque.

Vous serez uniquement autorisés à vous concentrer sur vos sensations internes de vos battements cardiaques, mais ne mesurez pas votre rythme cardiaque par d'autres moyens (ex. vérification du pouls au poignet ou au cou).
        """
    texts[
        "pulseTutorial1"
    ] = "Veuillez placer l'oxymètre de pouls sur votre index. Utilisez votre main non-dominante comme illustré sur ce schéma."

    texts[
        "pulseTutorial2"
    ] = "Si vous pouvez sentir vos battements de coeur quand vous portez l'oxymètre de pouls sur votre index, essayez de le placer sur un autre doigt."

    texts[
        "pulseTutorial3"
    ] = "Vous pouvez essayer différentes configurations jusqu'à ce que vous trouviez le doigt qui provoque le moins de sensations de battements cardiaques."

    texts[
        "pulseTutorial4"
    ] = "Veuillez entrer le numéro du doigt correspondant au doigt sur lequel vous avez décidé de placer l'oxymètre de pouls."

    texts[
        "Tutorial2"
    ] = "Quand vous voyez cette icône, essayez de vous concentrer sur vos battements cardiaques durant 5 secondes. Essayez de ne pas bouger, car nous enregistrons votre pouls durant cette période."

    moreResp = "flèche vers le HAUT" if device == "keyboard" else "clic DROIT"
    lessResp = "flèche vers le BAS" if device == "keyboard" else "clic GAUCHE"
    texts[
        "Tutorial3_icon"
    ] = """Après cette période d'écoute du coeur, vous verrez la même icône and entendrez une série de bips."""
    texts[
        "Tutorial3_responses"
    ] = f"""Aussi rapidement et précisément possible, vous écouterez ces bips et déciderez s'ils sont plus rapides ({moreResp}) ou plus lents ({lessResp}) que votre propre rythme cardiaque.

Les bips seront TOUJOURS plus lents ou plus rapides que votre coeur. Veuillez faire une estimation, même si vous n'est pas sûr."""

    if exteroception is True:
        texts[
            "Tutorial3bis"
        ] = """Pour certains essais, au lieu de voir une icône de coeur, vous verrez une icône d'écoute. Vous devrez alors écouter une première série de bips, au lieu de votre coeur."""

        texts[
            "Tutorial3ter"
        ] = f"""Après ce premier bip, vous verrez l'icône de réponse apparaître, et une seconde série de bip sera joué.

Aussi rapidement et précisément possible, vous entendrez ces bips et déciderez s'ils sont plus rapides ({moreResp}) ou plus lents ({lessResp}) que la première série de bips.

La seconde série de bips sera TOUJOURS plus lente ou rapide que la première série. Veuillez faire une estimation, même si vous n'êtes pas sûr."""

    texts[
        "Tutorial4"
    ] = """Une fois que vous avez donné votre réponse, il vous sera également demandé d'estimer votre degré de confiance dans votre réponse.

Ici, le score maximum signifie que vous êtes totalement certain de votre choix, et le score minimum signifie que vous devinez.

Vous devez utiliser la souris pour sélectionner votre score"""

    texts[
        "Tutorial5"
    ] = """Cette séquence sera répétée durant la tâche.

Par moment la tâche peut être très difficile ; la différence entre votre propre rythme cardiaque et les bips présentés peut être très petite.

Cela signifie que vous devez essayer d'utiliser toute la longueur de l'échelle de confiance pour refléter votre incertitude subjective sur chaque essai.

Comme la difficulté de la tâche évolue avec le temps, il est rare que vous soyez totalement confiant ou totalement incertain."""

    texts[
        "Tutorial6"
    ] = """Ceci conclut le tutoriel. Si vous avez des questions, veuillez les poser maintenant à l'expérimentateur.
Sinon, vous pouvez continuer avec la tâche principale."""

    return texts



# Author: Nicolas Legrand <nicolas.legrand@cas.au.dk>

import os
from typing import Any, Dict, Optional

import numpy as np
import pandas as pd
import pkg_resources  # type: ignore
import serial
from systole import serialSim
from systole.recording import Oximeter

from cardioception.HRD.languages import danish, danish_children, english, french


def getParameters(
    participant: str = "SubjectTest",
    session: str = "001",
    serialPort: str = "COM3",
    setup: str = "behavioral",
    stairType: str = "psi",
    exteroception: bool = True,
    catchTrials: float = 0.0,
    nTrials: int = 120,
    device: str = "mouse",
    screenNb: int = 0,
    fullscr: bool = True,
    nBreaking: int = 20,
    resultPath: Optional[str] = None,
    language: str = "english",
    systole_kw: dict = {},
):
    """Create Heart Rate Discrimination task parameters.

    Many task parameters, aesthetics, and options are controlled by the
    parameters dictionary defined herein. These are intended to provide
    flexibility and modularity to the task. In many cases, unique versions of the
    task (e.g., with or without confidence ratings or choice feedback) can be
    created simply by changing these parameters, with no further interaction
    with the underlying task code.

    Parameters
    ----------
    device :
        Select how the participant provides responses. Can be `'mouse'` or `'keyboard'`.
    exteroception :
        If `True`, the task will include an exteroceptive (half of the trials).
    fullscr :
        If `True`, activate full-screen mode.
    language :
        The language used for the instruction. Can be `"english"`, `"danish"` or
        `"danish_children"` (a slightly simplified danish version), or `"french"`.
    nBreaking :
        Number of trials to run before the break.
    nStaircase :
        Number of staircases to use per condition (exteroceptive and
        interoceptive).
    nTrials :
        The number of trials to run (UpDown and psi staircase).
        .. note::
           This number indicates the total number of trials that will be presented
           during the experiment. If `nTrials=50` and `exteroception=False`, the task
           contains 50 interoceptive trials. If `nTrials=50` and `exteroception=True`,
           the task contains 25 interoceptive trials and 25 exteroceptive trials.
    participant :
        Subject ID. The default is 'Participant'.
    catchTrials :
        Ratio of Psi trials allocated to extreme values (+20 or -20 bpm with some
        jitter) to control for a range of stimuli presented. Default to `0.0` (no catch
        trials). If not `0.0`, recommended value is `0.2`.
    resultPath :
        Where to save the results.
    screenNb :
        Screen number. Used to parametrize py:func:`psychopy.visual.Window`. Defaults
        to `0`.
    serialPort:
        The USB port where the pulse oximeter is plugged. Should be written as a string
        e.g. `"COM3"` for USB ports on Windows.
    session :
        Session number. Default to '001'.
    setup :
        Context of oximeter recording. `"ehavioral"` will be recorded through a Nonin
        pulse oximeter and `"test"` will use a pre-recorded pulse time series (for
        testing only).
    stairType :
        Staircase type. Can be "psi" or "updown". The default is set to "psi".
    systole_kw :
        Additional keyword arguments for :py:class:`systole.recorder.Oxmeter`.

    Attributes
    ----------
    confScale :
        The range of the confidence rating scale.
    device :
        The device used for response and rating scale. Can be `"keyboard"` or
        `"mouse"`.
    HRcutOff :
        Cut off for extreme heart rate values during recording.
    ExteroCondition :
        If `True`, the task includes an exteroceptive (half of the trials).
    isi :
        Range of the inter-stimulus interval (seconds). Should be in the form of (low,
        high). At each trial, the value is generated using a uniform distribution
        between these two values. The default is set to `(0.25, 0.25)` so the value is
        fixed at `0.25`.
    labelsRating :
        The labels of the confidence rating scale.
    lambdaExtero :
        (3d) Posterior estimate of the psychophysics function parameters (slope and
        threshold) across trials for the exteroceptive condition.
    lambdaIntero :
        (3d) Posterior estimate of the psychophysics function parameters (slope and
        threshold) across trials for the interoceptive condition.
    listenLogo, heartLogo : Psychopy visual instance
        Image used for the inference and recording phases, respectively.
    maxRatingTime :
        The maximum time for a confidence rating (in seconds).
    minRatingTime :
        The minimum time before a rating can be provided during the confidence
        rating (in seconds).
    monitor :
        The monitor used to present the task (Psychopy parameter).
    nBreaking :
        Number of trials to run before the break.
    nConfidence :
        The number of trials with feedback during the tutorial phase (no
        feedback).
    nFeedback :
        The number of trials with feedback during the tutorial phase (no
        confidence rating).
    nFinger :
        The finger number ("1", "2", "3", "4" or "5") where the participant
        decided to place the pulse oximeter (if relevant).
    nTrials :
        The number of trials to run (UpDown and psi staircase).
        .. note::
           This number indicates the total number of trials that will be presented
           during the experiment. If `nTrials=50` and `exteroception=False`, the task
           contains 50 interoceptive trials. If `nTrials=50` and `exteroception=True`,
           the task contains 25 interoceptive trials and 25 exteroceptive trials.
    participant :
        Subject ID. The default is 'Participant'.
    path :
        The task working directory.
    response_keys :
        A dictionary listing the possible response key for Faster/More and Slower/Less
        trials. The default is `"up"`/`"down"`. Only relevant if `device=="keyboard"`.
    resultPath :
        Where to save the results.
    serial :
        The serial port is used to record the PPG activity.
    screenNb :
        The screen number (Psychopy parameter). The default is set to 0.
    signal_df :
        Dataframe where the pulse signal recorded during the interoception
        condition will be stored.
    stairCase :
        The staircase instances for 'psi' and 'UpDown'. Each entry contains
        a dictionary for 'Intero' and 'Extero conditions' (if relevant).
    staircaseType :
        Vector indexing stairce type (`'UpDown'`, `'psi'`, `'psiCatchTrial'`).
    startKey :
        The key to press to start the task and go to the next steps.
    respMax :
        The maximum time for decision (in seconds).
    results :
        The result directory.
    session :
        Session number. Default to '001'.
    setup :
        The context of recording. Can be `'behavioral'` or `'test'`.
    texts :
        Long text elements.
    textSize :
        Scaling parameter for text size.
    triggers :
        Dictionary {str, callable or None}. The function will be executed
        before the corresponding trial sequence. The default values are
        `None` (no trigger sent).
        * `"trialStart"`
        * `"trialStop"`
        * `"listeningStart"`
        * `"listeningStop"`
        * `"decisionStart"`
        * `"decisionStop"`
        * `"confidenceStart"`
        * `"confidenceStop"`
    win :
        The window in which to draw objects.

    Notes
    -----
    When using the `behavioral` setup, triggers will be sent to the PPG  recording. The
    trigger channel is coding for different events during the task as follows:
    - Trial start: 1
    - recording trigger: 2
    - sound trigger : 3
    - rating trigger: 4
    - end trigger: 5
    All these events, except the trial start, have also their time stamps encoded in the
    behavioural results data frame.

    """
    from psychopy import data, event, visual

    parameters: Dict[str, Any] = {}
    parameters["ExteroCondition"] = exteroception
    parameters["device"] = device
    if parameters["device"] == "keyboard":
        parameters["confScale"] = [1, 7]
    parameters["labelsRating"] = ["Guess", "Certain"]
    parameters["screenNb"] = screenNb
    parameters["monitor"] = "testMonitor"
    parameters["nFeedback"] = 5
    parameters["nConfidence"] = 8
    parameters["respMax"] = 5
    parameters["minRatingTime"] = 0.5
    parameters["maxRatingTime"] = 5
    parameters["isi"] = (0.25, 0.25)
    parameters["startKey"] = "space"
    parameters["response_keys"] = {"More": "up", "Less": "down"}
    parameters["nTrials"] = nTrials
    parameters["nBreaking"] = nBreaking
    parameters["lambdaIntero"] = []  # Save the history of lambda values
    parameters["lambdaExtero"] = []  # Save the history of lambda values
    parameters["nFinger"] = None
    parameters["signal_df"] = pd.DataFrame([])  # Physiological recording
    parameters["results_df"] = pd.DataFrame([])  # Behavioral results

    # Set default path /Results/ 'Subject ID' /
    parameters["participant"] = participant
    parameters["session"] = session
    parameters["path"] = os.getcwd()
    if resultPath is None:
        parameters["resultPath"] = parameters["path"] + "/data/" + participant + session
    else:
        parameters["resultPath"] = None
    # Create Results directory if not already exists
    if not os.path.exists(parameters["resultPath"]):
        os.makedirs(parameters["resultPath"])

    # Store posterior in a dictionary
    parameters["staircaisePosteriors"] = {}
    parameters["staircaisePosteriors"]["Intero"] = []
    if exteroception is True:
        parameters["staircaisePosteriors"]["Extero"] = []

    nCatch = int(parameters["nTrials"] * catchTrials)
    nStaircase = parameters["nTrials"] - nCatch

    # Vector encoding the staircase type
    if stairType == "psi":
        sc = np.array(["psi"] * nStaircase)
    elif stairType == "updown":
        sc = np.array(["updown"] * nStaircase)
    else:
        raise ValueError("stairType should be 'psi' or 'updown'")

    # Create and randomize condition vectors separately for each staircase
    if exteroception is True:
        # Create a modality vector containing nTrials/2 Intero and Extero conditions
        parameters["Modality"] = np.hstack(
            [np.array(["Extero", "Intero"] * int(parameters["nTrials"] / 2))]
        )
    elif exteroception is False:
        # Create a modality vector containing nTrials/2 Intero and Extero conditions
        parameters["Modality"] = np.array(["Intero"] * int(parameters["nTrials"]))
    else:
        raise ValueError("exteroception should be a boolean")

    # Vector encoding the type of trial (psi, up/down or catch)
    parameters["staircaseType"] = np.hstack(
        [
            sc,
            np.array(["CatchTrial"] * int((parameters["nTrials"] * catchTrials))),
        ]
    )

    # Shuffle all trials
    shuffler = np.random.permutation(parameters["nTrials"])
    parameters["Modality"] = parameters["Modality"][shuffler]
    parameters["staircaseType"] = parameters["staircaseType"][shuffler]

    # Default parameters for the basic staircase are set here. Please see
    # PsychoPy Staircase Handler Documentation for full options. By default,
    # the task implements a staircase using Psi method.
    # If UpDown is selected, 1 or 2 interleaved staircases are used (see
    # options in parameters dictionary), one is initialized 'high' and the other
    # 'low'.
    parameters["stairCase"] = {}

    if stairType == "updown":
        conditions = [
            {
                "label": "low",
                "startVal": -40.5,
                "nUp": 1,
                "nDown": 1,
                "stepSizes": [20, 12, 12, 7, 4, 3, 2, 1],
                "stepType": "lin",
                "minVal": -40.5,
                "maxVal": 40.5,
            },
            {
                "label": "high",
                "startVal": 40.5,
                "nUp": 1,
                "nDown": 1,
                "stepSizes": [20, 12, 12, 7, 4, 3, 2, 1],
                "stepType": "lin",
                "minVal": -40.5,
                "maxVal": 40.5,
            },
        ]
        parameters["stairCase"]["Intero"] = data.MultiStairHandler(
            conditions=conditions, nTrials=parameters["nTrials"]
        )

    elif stairType == "psi":
        parameters["stairCase"]["Intero"] = data.PsiHandler(
            nTrials=nTrials,
            intensRange=[-50.5, 50.5],
            alphaRange=[-50.5, 50.5],
            betaRange=[0.1, 25],
            intensPrecision=1,
            alphaPrecision=1,
            betaPrecision=0.1,
            delta=0.02,
            stepType="lin",
            expectedMin=0,
        )

    if exteroception is True:
        if stairType == "updown":
            conditions = [
                {
                    "label": "low",
                    "startVal": -40.5,
                    "nUp": 1,
                    "nDown": 1,
                    "stepSizes": [20, 12, 12, 7, 4, 3, 2, 1],
                    "stepType": "lin",
                    "minVal": -40.5,
                    "maxVal": 40.5,
                },
                {
                    "label": "high",
                    "startVal": 40.5,
                    "nUp": 1,
                    "nDown": 1,
                    "stepSizes": [20, 12, 12, 7, 4, 3, 2, 1],
                    "stepType": "lin",
                    "minVal": -40.5,
                    "maxVal": 40.5,
                },
            ]
            parameters["stairCase"]["Extero"] = data.MultiStairHandler(
                conditions=conditions, nTrials=parameters["nTrials"]
            )

        elif stairType == "psi":
            parameters["stairCase"]["Extero"] = data.PsiHandler(
                nTrials=nTrials,
                intensRange=[-50.5, 50.5],
                alphaRange=[-50.5, 50.5],
                betaRange=[0.1, 25],
                intensPrecision=1,
                alphaPrecision=1,
                betaPrecision=0.1,
                delta=0.02,
                stepType="lin",
                expectedMin=0,
            )

    parameters["setup"] = setup
    if setup == "behavioral":
        # PPG recording
        port = serial.Serial(serialPort)
        parameters["oxiTask"] = Oximeter(
            serial=port, sfreq=75, add_channels=1, **systole_kw
        )
        parameters["oxiTask"].setup().read(duration=1)
    elif setup == "test":
        # Use pre-recorded pulse time series for testing
        port = serialSim()
        parameters["oxiTask"] = Oximeter(
            serial=port, sfreq=75, add_channels=1, **systole_kw
        )
        parameters["oxiTask"].setup().read(duration=1)

    ##############
    # Load texts #
    ##############
    if language == "english":
        parameters["texts"] = english(
            device=device, setup=setup, exteroception=exteroception
        )
    elif language == "danish":
        parameters["texts"] = danish(
            device=device, setup=setup, exteroception=exteroception
        )
    elif language == "danish_children":
        parameters["texts"] = danish_children(
            device=device, setup=setup, exteroception=exteroception
        )
    elif language == "french":
        parameters["texts"] = french(
            device=device, setup=setup, exteroception=exteroception
        )

    # Open window
    if parameters["setup"] == "test":
        fullscr = False
    parameters["win"] = visual.Window(
        monitor=parameters["monitor"],
        screen=parameters["screenNb"],
        fullscr=fullscr,
        units="height",
    )
    parameters["win"].mouseVisible = False

    ###############
    # Image loading
    ###############
    if parameters["setup"] in ["test", "behavioral"]:
        parameters["pulseSchema"] = visual.ImageStim(
            win=parameters["win"],
            units="height",
            image=pkg_resources.resource_filename(__name__, "Images/pulseOximeter.png"),
            pos=(0.0, 0.0),
        )
        parameters["pulseSchema"].size *= 0.2
        parameters["handSchema"] = visual.ImageStim(
            win=parameters["win"],
            units="height",
            image=pkg_resources.resource_filename(__name__, "Images/hand.png"),
            pos=(0.0, -0.08),
        )
        parameters["handSchema"].size *= 0.15

    parameters["listenLogo"] = visual.ImageStim(
        win=parameters["win"],
        units="height",
        image=pkg_resources.resource_filename(__name__, "Images/listen.png"),
        pos=(0.0, 0.0),
    )
    parameters["listenLogo"].size *= 0.08

    parameters["heartLogo"] = visual.ImageStim(
        win=parameters["win"],
        units="height",
        image=pkg_resources.resource_filename(__name__, "Images/heartbeat.png"),
        pos=(0.0, 0.0),
    )
    parameters["heartLogo"].size *= 0.04
    parameters["textSize"] = 0.04
    parameters["HRcutOff"] = [40, 120]
    if parameters["device"] == "keyboard":
        parameters["confScale"] = [1, 10]
    elif parameters["device"] == "mouse":
        parameters["myMouse"] = event.Mouse()

    return parameters



# Author: Nicolas Legrand <nicolas.legrand@cas.au.dk>

# Running short blocks of the task for testing.
# Shoult not be used for data acquisition

import shutil
import unittest
from unittest import TestCase

from cardioception.HBC.parameters import getParameters
from cardioception.HBC.task import run


class TestHBC(TestCase):
    def test_parameters(self):
        """Test get_parameters function"""
        # Get parameters
        parameters = getParameters(setup="test")
        parameters["win"].close()
        shutil.rmtree(parameters["resultPath"])

    def test_run(self):
        """Test run function"""
        # Get parameters
        parameters = getParameters(setup="test", taskVersion="test")

        run(parameters)

        parameters["win"].close()
        shutil.rmtree(parameters["resultPath"])


if __name__ == "__main__":
    unittest.main(argv=["first-arg-is-ignored"], exit=False)



# Author: Nicolas Legrand <nicolas.legrand@cas.au.dk>

# Running short blocks of the task for testing.
# Shoult not be used for data acquisition

import shutil
import unittest
from unittest import TestCase

import numpy as np
from psychopy import prefs

prefs.hardware["audioLib"] = ["pygame"]

from cardioception.HRD.parameters import getParameters
from cardioception.HRD.task import run


class TestHRD(TestCase):
    def test_parameters(self):
        """Test parameters function"""

        parameters = getParameters(
            setup="test", nTrials=80, exteroception=True, stairType="psi"
        )
        parameters["win"].close()

        assert len(parameters["Modality"]) == 80
        assert sum(parameters["Modality"] == "Extero") == 40
        assert len(parameters["staircaseType"]) == 80
        assert np.all(parameters["staircaseType"] == "psi")

        parameters = getParameters(
            setup="test",
            nTrials=4,
            exteroception=True,
            stairType="updown",
            catchTrials=0.2,
        )
        parameters["win"].close()
        shutil.rmtree(parameters["resultPath"])

        assert sum(parameters["Modality"] == "Intero") == 2
        assert len(parameters["Modality"]) == 4
        assert len(parameters["staircaseType"]) == 4
        assert sum(parameters["staircaseType"] == "updown") == 4

    def test_run(self):
        """Test run function"""

        # VErsion 1
        parameters = getParameters(
            setup="test",
            nTrials=4,
            exteroception=True,
            stairType="psi",
            catchTrials=0.5,
        )
        parameters["nConfidence"] = 1
        parameters["nFeedback"] = 1

        run(parameters, confidenceRating=True, runTutorial=True)
        parameters["win"].close()
        shutil.rmtree(parameters["resultPath"])

        # Version 2
        parameters = getParameters(
            setup="test",
            nTrials=4,
            exteroception=False,
            stairType="updown",
            device="keyboard",
            catchTrials=0.0,
        )

        run(parameters, confidenceRating=True, runTutorial=False)
        parameters["win"].close()
        shutil.rmtree(parameters["resultPath"])


if __name__ == "__main__":
    unittest.main(argv=["first-arg-is-ignored"], exit=False)



# Author: Nicolas Legrand <nicolas.legrand@cas.au.dk>
# Adapted from: https://stackoverflow.com/questions/33879523/python-how-can-i-generate-a-wav-file-with-beeps
# This script generates the BPM stimuli (wav files) used by the task.
# Not called by the actual task, but included for reproducibility.

import math
import struct
import wave

import numpy as np


def append_silence(audio: list, duration_milliseconds: int = 500):
    """Add silence to the signal.

    Parameters
    ----------
    audio : list
        The signal where the sine should be added.
    duration_milliseconds : int
        The sine length.

    Returns
    -------
    audio : list
        The signal with the added silence.
    """
    num_samples = duration_milliseconds * (sample_rate / 1000.0)

    for x in range(int(num_samples)):
        audio.append(0.0)

    return audio


def append_sinewave(
    audio: list, freq: int = 440, duration_milliseconds: int = 200, volume: float = 1.0
):
    """Add sinewave to the signal.

    Parameters
    ----------
    audio : list
        The signal where the sine should be added.
    freq : int
        The sine frequency.
    duration_milliseconds : int
        The sine length.
    volume : float
        The sine amplitude.

    Returns
    -------
    audio : array
        The signal with the added sound.
    """
    num_samples = duration_milliseconds * (sample_rate / 1000.0)

    for x in range(int(num_samples)):
        audio.append(volume * math.sin(2 * math.pi * freq * (x / sample_rate)))

    return audio


def save_wav(audio: list, file_name: str):
    """Save the audio signal as wav file.

    Parameters
    ----------
    audio : list
        The signal where the sine should be added.
    freq : int
        The sine frequency.
    duration_milliseconds : int
        The sine length.
    volume : float
        The sine amplitude.
    """
    # Open up a wav file
    wav_file = wave.open(file_name, "w")

    # wav params
    nchannels, sampwidth = 1, 2

    # 44100 is the industry standard sample rate
    nframes = len(audio)
    wav_file.setparams(
        (nchannels, sampwidth, sample_rate, nframes, "NONE", "not compressed")  # type: ignore
    )

    for sample in audio:
        wav_file.writeframes(struct.pack("h", int(sample * 32767.0)))

    wav_file.close()


# Generate wav files for frequencies between 15 and 200 beats per minutes
for bpm in np.arange(15, 200, 0.5):
    rr = (60000 / bpm) - 200
    audio: list = []
    sample_rate = 44100.0

    # Create
    beats = 5
    while beats > 0:
        # Sound
        audio = append_sinewave(audio, volume=0.5)
        audio = append_silence(audio, duration_milliseconds=rr)
        beats -= 1

    save_wav(audio, "Sounds/" + str(bpm) + ".wav")



from .languages import danish, danish_children, english, french
from .parameters import getParameters
from .task import (
    confidenceRatingTask,
    responseDecision,
    run,
    trial,
    tutorial,
    waitInput,
)

__all__ = [
    "getParameters",
    "confidenceRatingTask",
    "responseDecision",
    "run",
    "trial",
    "tutorial",
    "waitInput",
    "english",
    "danish",
    "danish_children",
    "french",
]



from typing import TYPE_CHECKING

from .reports import preprocessing, report

if not TYPE_CHECKING:
    from .HBC import *  # noqa
    from .HRD import *  # noqa

__all__ = [
    "preprocessing",
    "report",
]

__version__ = "0.5.0"



# Copyright (C) 2020 Nicolas Legrand
import os
import codecs
from setuptools import find_packages, setup

PROJECT_ROOT = os.path.dirname(os.path.realpath(__file__))
REQUIREMENTS_FILE = os.path.join(PROJECT_ROOT, "requirements.txt")

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

def get_requirements():
    with codecs.open(REQUIREMENTS_FILE) as buff:
        return buff.read().splitlines()

DESCRIPTION = "Cardioception Python Package"
LONG_DESCRIPTION = """Measuring interoceptive performance with Psychopy.
"""

DISTNAME = "cardioception"
MAINTAINER = "Nicolas Legrand"
MAINTAINER_EMAIL = "nicolas.legrand@cas.au.dk"
VERSION = "0.5.0"

if __name__ == "__main__":

    setup(
        name=DISTNAME,
        author=MAINTAINER,
        author_email=MAINTAINER_EMAIL,
        maintainer=MAINTAINER,
        maintainer_email=MAINTAINER_EMAIL,
        description=DESCRIPTION,
        long_description=LONG_DESCRIPTION,
        license=read("LICENSE"),
        version=VERSION,
        url="https://github.com/LegrandNico/Cardioception",
        install_requires=get_requirements(),
        include_package_data=True,
        packages=find_packages(),
        package_data={
            "cardioception.HBC": ["*.wav", "*.png"],
            "cardioception.HRD": ["*.wav", "*.png"],
            "cardioception.notebooks": ["*.ipynb"],
        },
    )
